#! /usr/bin/env python
#
# 
# This file is part of libRadtran.
# Copyright (c) 2010 by Jonas Kylling.
# 
# libRadtran is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# libRadtran is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with libRadtran.  If not, see <http://www.gnu.org/licenses/>.

"""
GUI for uvspec

To run the GUI you need
    python 2.5 or 2.6. Older versions might work
    wxPython 2.8

Stdout and stderr are redirected to .UVSPEC_ERR. This can be turned on and off by setting redirect to True or False.

Plotting
The GUI should be able to plot most output file made with uvspec. Special care
is needed when plotting output files with the umu, zout and phi options. The GUI
uses the number of values in umu, phi and zout to parse the output file. 
If you want to plot an output file from a previous run, you have to open the inputfile
your created the output file with(or type in the same number of values in umu, phi
and zout).

"""

import os
import sys
import re
import time
import threading
import subprocess
import itertools
import pickle


try:
    import wx
    import wx.lib.scrolledpanel
    from wx.lib import plot
except:
    print "\nError occured while importing wx!"
    print "Quitting..."
    raise
    
if wx.__version__ != "2.8.9.1":
    print "\nYou do not have the same wx version as the developer! The GUI might not work as expected:-)\n"
    print "Your version of wx is %s." % wx.__version__

try:
    import widgets
    import create
except:
    print "\nError occured while importing widgets or create!"
    print "Quitting..."
    raise

#----------------------------------------Globals----------------------------------------------------
__version__ = version = "1.0"


redirect = False # turn on off redirection of stdout and stderr
redirect_file = ".UVSPEC_ERR" # stdout and stderr are redirected to this file

window_size = (1000, 450)
window_title = "GUI for libRadtran"

DocWindow = True
ToolTips = False


def pr(path):
    global libRadtran_GUI_path
    return os.path.join(libRadtran_GUI_path, path)

def pe(path):
    global libRadtran_examples_path
    return os.path.join(libRadtran_examples_path, path)

widgets.pr = pr
widgets.pe = pe
create.pr = pr
create.pe = pe


def Load_Documentation():
    """
    Load the documentation from GUI_DOC
    The documentation is saved as a dictionary
    """
    try:
        doc_file = open(pr("resources/GUI_DOC"), "r")
        Documentation = pickle.load(doc_file)
        doc_file.close()
    except IOError:
        print "Failed to load doc file GUI_DOC.\a"
        Documentation = {}
    # dictionary with needed keys that are normally not in GUI_DOC
    not_set = {"Input File": "Input File",
               "Output File": "Output File",
               "Plot_Out":"Plot a data file"}
    for name, value in not_set.items():
        try:
            Documentation[name]
        except KeyError:
            Documentation[name] = value
        
        Documentation["Output File"] = "Output File"
        Documentation["Plot_Out"] = "Plot a data file"

    return Documentation

if ToolTips:
    Documentation = Load_Documentation()

def SplashScreen():
    """
    Display a splashscreen
    """
    try:
        if not os.path.exists(pr("resources/splash.png")):
            print "Could not find splash.png"
            return
        image = wx.Image(pr("resources/splash.png"), wx.BITMAP_TYPE_ANY)
        image = image.ConvertToBitmap()
        splashstyle = wx.SPLASH_CENTRE_ON_SCREEN|wx.SPLASH_NO_TIMEOUT
        style = wx.SIMPLE_BORDER#|wx.STAY_ON_TOP
        SplashWindow = wx.SplashScreen(image, splashstyle, -1, parent=None)
        return SplashWindow
    except Exception, error:
        print "Could not create splash screen.'"
        print error

FileDialog = widgets.FileDialog
ErrorMessage = widgets.ErrorMessage


class PlotPanel(wx.Panel):
    def __init__(self, parent, info="Info", format="*.*", choices=[], **kwargs):
        wx.Panel.__init__(self, parent, style=wx.BORDER_SUNKEN)
        
        # Grammar is important
        if info[0].capitalize() in ["A", "E" "I", "O", "U", "Y"]: information = "Plot an %s" % info
        else: information = "Plot a %s" % info
        
        self.Button = wx.Button(self, label="Plot")
        self.FileInfo = widgets.OpenButton(self, name=info, format=format, without_clear_button=True)
        self.FileInfo.end_init()
        self.Choices = widgets.PlotChoices(self, choices=choices)
        self.VarList = widgets.PlotXAgainstY(self, title="output_user")

        main_sizer = wx.BoxSizer(wx.VERTICAL)
        info_sizer = wx.BoxSizer(wx.HORIZONTAL)

        info_sizer.Add(self.Button, 0, wx.RIGHT|wx.CENTER|wx.TOP|wx.BOTTOM, border=5)
        info_sizer.Add(self.FileInfo, 1, wx.CENTER|wx.RIGHT|wx.EXPAND, border=5)
        info_sizer.Add(self.Choices)

        main_sizer.Add(info_sizer, 0, wx.EXPAND)
        main_sizer.Add(self.VarList, 1)

        self.Button.Bind(wx.EVT_BUTTON, self.OnPlot)
        Options["output_user"].Bind(wx.EVT_TEXT, self.UpdateVars)
        Options["output_user"].SetValue = widgets.AttachFunction(self.UpdateVars, Options["output_user"].SetValue)

        self.SetSizer(main_sizer)
    Has_Changed = False
    def end_init(self): pass
    def Clear(self):
        self.FileInfo.Clear()
        self.Choices.Clear()
        self.VarList.Clear()
    def UpdateVars(self, event):
        tmp = Options["output_user"].GetValue()
        if tmp:
            self.VarList.UpdateVars(tmp.strip().split())
            self.VarList.Enable(True)
        else:
            self.VarList.Enable(False)
        if event:
            event.Skip()
    def OnPlot(self, event):
        """Setting everything up for plotting"""
        path = self.FileInfo.GetValue()
        output_user = Options["output_user"].GetValue()
	
        xtitle = "Wavelength (nm)"
        ytitle = "Radiation (quantity and units depends on user input)"

        if not path:
            msg = "You have to chose an output file to plot."
            print msg
            ErrorMessage(msg)
            return
        if not os.path.exists(path):
            msg = "Could not find '%s'." % (path)
            print msg
            ErrorMessage(msg)
            return
        if not (any([box.IsChecked() for box in self.Choices])) and not output_user:
            msg = "You have to check at least one of the checkboxes."
            print msg
            ErrorMessage(msg)
            return
        varlist = self.VarList.GetValue()
        if output_user and (not any([i[1] for i in varlist]) or not any(i[2] for i in varlist)):
            msg = "You have to mark at least on x and y variable to plot."
            print msg
            ErrorMessage(msg)
            return
        
        if output_user:
            data, min_max = self.ReadVars(path, self.VarList.GetValue())
            y_list = []
            for i in varlist:
                if i[1]:
                    xtitle = i[0].title()
                if i[2]: y_list.append(i[0])
            ytitle = ", ".join(y_list).title()
        else:
            try:
                data, min_max = self.ReadData(path)
            except TypeError: # catch error when ReadData returns None
                return
        #self.ManyPlot(data, min_max)
        widgets.PlotWindow(self, data, MinMax=min_max, title="Radiation Plot", xtitle=xtitle,
                    ytitle=ytitle)
        
    def ManyPlot(self, data, min_max):
        """ Plot all the points in a seperate window. Just for debug. """
        count = 1
        for points in data:
            widgets.PlotWindow(self, (points,), MinMax=min_max, title="Plot %i" % count, xtitle="Wavelength (nm)",
                       ytitle="Radiation")
            count += 1
            
    def ReadData(self, path):
        try:
            f = open(path, "r")
            data = f.readlines()
            f.close()
        except IOError, error:
            ErrorMessage("Could not open file %s." % (path))
            print error
            
        try:
            for pos in xrange(len(data)):
                data[pos] = data[pos].split()
                if data[pos][0].startswith("#"): continue
                data[pos] = [float(num) for num in data[pos]]
        except ValueError, error:
            ErrorMessage("Invalid characters in %s. Your file might be corrupted." % (path))
            print error

        if not data:
            ErrorMessage("Empty file.")
        
        if (Options["umu"].GetValue() and Options["phi"].GetValue()):
            return self.ReadBlocks(data)
        elif any([box.IsChecked() for box in self.Choices]):
            return self.ReadColumns(data)
        
    def ReadBlocks(self, data):
        """ Read an output file in blocks """

        # structure of each block:
        # line with wavelength data
        # line with degrees
        # line with data
        # line with more data ...

        zout = Options["zout"].GetValue()
        if zout:
            legends = zout.split()
            zout = len(legends)
        if not zout:
            zout = 1 # defaults to one
            legends = ["boa"]

        height = len(Options["umu"].GetValue().split()) + 2 # returns the length of the data
        length =  len(Options["phi"].GetValue().split()) # returns the number of datafields we need.
                                                                        # the two first data fields of each line must be skipped
        blocks = [ data[pos:height + pos] for pos in xrange(0, len(data), height) ] # divide the file into separate blocks
        
        graphs = [] # contains all the PolyLines
        x_bounds = [] # contains min, max of the x axis
        y_bounds = []

        #Parsing of the irrdaiances
        for box in self.Choices:
            if box.IsChecked(): # just test if any of the boxes are checked
                tmp_graphs, min_max = self.ReadColumns([block[0] for block in blocks])
                x_bounds.extend(min_max[0])
                y_bounds.extend(min_max[1])
                graphs.extend(tmp_graphs)
                break

        #Parsing of the radiances
        if self.Choices["radiances"].IsChecked():
            points = [[] for i in xrange((height - 2) * length * zout)] # initilaize a list wich will contain the points
            legend_list =  [[] for i in xrange((height - 2) * length * zout)]# legend for each graph, the legends is "boa umu phi"
            zout_val = 0
            try:
                for block in blocks:
                    wave_length = block[0][0]
                    x_bounds.append(wave_length)
                    c = -2 # explained below
                    for line in block[2:]: # skip the two first lines
                        # pos starts at 2 and the first value is at line[2]
                        # the value on the line is stored in out[i + c]
                        # since out starts at zero h has to be -2 and
                        # incremented with the length for every iteration
                        # over each line in each block. 
                        for pos in xrange(2, length + 2): # skip the two first fields
                            points[zout_val + c + pos].append((wave_length, line[pos]))
                            y_bounds.append(line[pos])
                            if not legend_list[zout_val + c + pos]:
                                legend_list[zout_val + c + pos] = legends[zout_val/(length * (height - 2))],  line[0], block[1][pos - 2]
                        c += length
                    zout_val += length * (height -2)
                    if zout_val == zout * length * (height -2): zout_val = 0
            except IndexError, error:
                print error
                ErrorMessage("An error occured while reading the file. Please control that your input file and output file cohere.")
                return
                
            colour = widgets.BeautyColour(len(points))
            graphs.extend([ plot.PolyLine(points[i], legend="%5s %5s %5s" % legend_list[i], colour=colour.next(), width=1) for i in xrange(len(points)) ])
        
        if not graphs:
            ErrorMessage("You have to check at least one checkbox.")
            return
        return (graphs, ( (int(min(x_bounds)), int(max(x_bounds)),),
             (int(min(y_bounds)), int(max(y_bounds)),), ),
             )
            
    def ReadColumns(self, data):
        zout = Options["zout"].GetValue()
        if zout:
            legends = zout.split()
            zout = len(legends)
        if not zout:
            zout = 1 # defaults to one
            legends = ["boa"]

        styles = itertools.cycle((wx.SOLID, wx.DOT, wx.LONG_DASH, wx.SHORT_DASH,
                                      wx.HORIZONTAL_HATCH, wx.VERTICAL_HATCH))

        
        x_bounds = []
        y_bounds = []

        column_names = ["direct", "diff down", "upward"]
        
        column_num = 0 # The column of the out file to print
        columns = [] # list of the columns that will be plotted
        points = []

        legend_list = []
        style_list = []

        # initialize lists of various data
        for box in self.Choices:
            column_num += 1
            if box.IsChecked():
                for i in xrange(zout):
                    points.append([]); legend_list.append([]); style_list.append([])
                columns.append(column_num)

        # parse the data, lots of magic numbers
        c = 0
        for i in xrange(1, zout + 1):
            c += 1
            style = styles.next()
            for line in data:
                if c % zout == 0:
                    x_bounds.append(line[0])
                    for pos in xrange(len(columns)):
                        l_pos = pos + len(columns)*(i - 1) 
                        points[l_pos].append((line[0], line[columns[pos]]))
                        y_bounds.append(line[columns[pos]])
                        if not legend_list[l_pos]:
                            legend_list[l_pos] = "%s %s" % (legends[i - 1], column_names[columns[pos] - 1])
                            style_list[l_pos] = style
                c += 1

        return ([ plot.PolyLine(points[pos], legend=legend_list[pos],
                                colour=["RED", "GREEN", "BLUE"][columns[pos] - 1], width=2, style=style_list[pos]) for pos in xrange(len(points)) ],
                     (
                      ( int(min(x_bounds) ), int(max(x_bounds)) ),
                      ( int(min(y_bounds) ), int(max(y_bounds)) ),
                      )
                )
                # Max and min values of x and y axis
    def ReadVars(self, path, variables):
        c = 0
        
        x_var = None
        y_vars = []
        
        for title, x, y in variables:
            if x: x_var = title, c
            if y: y_vars.append((title, c))
            c += 1

        x_bounds = []
        y_bounds = []

        try:
            f = open(path, "r")
            data = f.readlines()
            f.close()
        except IOError, error:
            ErrorMessage("Could not open file %s." % (path))
            print error

        points = [[] for i in y_vars]
        try:
            for line in data:
                line = line.split()
                y_tmp = [float(line[y[1]]) for y in y_vars]
                x_tmp = float(line[x_var[1]])
            
                x_bounds.append(x_tmp)
                y_bounds.append(max(y_tmp))
                y_bounds.append(min(y_tmp))

                [points[i].append((x_tmp, y_tmp[i]))for i in xrange(len(y_tmp))]
        except IndexError:
            ErrorMessage("Malformed .OUT file.")
            

        colour = widgets.BeautyColour(len(points))
        return ([ plot.PolyLine(points[pos], legend=y_vars[pos][0],
                                width=1, colour=colour.next()) for pos in xrange(len(y_vars))],
                     (
                      ( int(min(x_bounds)), int(max(x_bounds))),
                      ( min(y_bounds), max(y_bounds) ),
                      ),
                )
        
    def SetValue(self, value):
        self.FileInfo.SetValue(value)
    def GetValue(self):
        return self.FileInfo.GetValue()


class Start(wx.Notebook):
    """The first panel
    
    Events for the buttons in this class are bound in the main class.
    The WelcomeText might be changed with html drawing for a better look
    """
    def __init__(self, parent):
        wx.Notebook.__init__(self, parent, )

        Home = wx.Panel(self, )
        Examples = wx.Panel(self)
        
        self.WelcomeText = wx.StaticText(Home, -1, """Welcome to the uvspec Graphical User Interface.
        
To start a new input file, push the New button.\n\n To open an old input file, push the Open button.

Mandatory options are in red, the rest are optional.        
        
Find more information about uvspec and \n libRadtran at www.libradtran.org.""", )#style=wx.ALIGN_CENTRE)

        self.New  = wx.Button(Home, -1, "New")
        self.Open = wx.Button(Home, -1, "Open")
        self.TryButton = wx.Button(Examples, -1, "Try", size=(200, -1))

        self.Examples = widgets.ExampleList(Examples, create.example_list)

        Logo = widgets.Logo(Home, -1, wx.Bitmap(pr("resources/esa.png")))

        main_sizer = wx.BoxSizer(wx.HORIZONTAL)
        common_sizer = wx.BoxSizer(wx.VERTICAL)
        button_sizer = wx.BoxSizer(wx.HORIZONTAL)
        example_sizer = wx.BoxSizer(wx.VERTICAL)

        button_sizer.Add(self.New, 0, wx.TOP|wx.RIGHT|wx.CENTER, border=20)
        button_sizer.Add(self.Open, 0, wx.TOP|wx.CENTER, border=20)
        
        common_sizer.Add(self.WelcomeText, 0, wx.CENTER)
        common_sizer.Add(button_sizer, 0, wx.CENTER)
        common_sizer.Add((-1, -1), 1)
        common_sizer.Add(Logo, 0, wx.ALIGN_RIGHT)
        
        example_sizer.Add(self.Examples, 1, wx.CENTER|wx.EXPAND|wx.BOTTOM, border=10)
        example_sizer.Add(self.TryButton, 0, wx.CENTER|wx.BOTTOM, border = 10)
        
        Home.SetSizer(common_sizer)
        Examples.SetSizer(example_sizer)
        
        self.AddPage(Home, "Home")
        self.AddPage(Examples, "Examples")
        

        
class Page(wx.Panel):
    """
    All pages in the notebook are instances of this class.
    """
    def __init__(self, parent, sizer_flag=wx.VERTICAL):
        """Optional sizer_flag, default is wx.VERTICAL"""
        wx.Panel.__init__(self, parent)
        self.Sizer = wx.BoxSizer(sizer_flag)

class NoteBook(wx.Notebook):
    """
    The funcitons CreatePage and CreatePanel is responsible for most of the the layout.
    """
    def __init__(self, parent):
        wx.Notebook.__init__(self, parent, -1, style=wx.NB_TOP)
        
    def AddPages(self, item_list, out):
        """A page should alway be added by calling this function"""
        for page in item_list:
            self.AddPage(*self.CreatePage(page, self, out))

    def CreatePage(self, item_list, parent, out):
        
        max_items = 8 # Max items per column
        
        #main_panel = wx.lib.scrolledpanel.ScrolledPanel(parent, style=wx.VSCROLL)
        main_panel = widgets.StupidScrolledWindow(parent, style=wx.VSCROLL)
        main_panel.Sizer = wx.BoxSizer(wx.HORIZONTAL)

        label = item_list[0][0]
        
        items = item_list[1:] # Everything except the panel label
        item_len = len(items)
        
        if item_len > max_items:
            times, left = divmod(item_len, 2)
            panel, option_dict = self.CreatePanel(main_panel, items[:times + left])
            out.update(option_dict)
            panel2, option_dict = self.CreatePanel(main_panel, items[times + left:])

            main_panel.Sizer.Add(panel, 1, wx.RIGHT, border=10)
            main_panel.Sizer.Add(panel2, 1,)
        else:
            panel, option_dict = self.CreatePanel(main_panel, items)
            
            main_panel.Sizer.Add(panel, 1, wx.EXPAND)
            
        out.update(option_dict)
        
        main_panel.SetSizer(main_panel.Sizer)
        main_panel.SetAutoLayout(1)
        main_panel.SetupScrolling()
        
        return main_panel, label
    
    def CreatePanel(self, parent, items):
            out = {}
            panel = Page(parent)

            largest_string = max([panel.GetTextExtent(string)[0] for string in [item[0]["name"] for item in items]])

            first = items[0][0]["name"]

            for item in items:
                if not redirect and 0: # don't clutter up the log files
                    print "  %s" % item[0]["name"], item[1].__name__  # REMOVE! Watch the progress when the GUI is initalizing
                item[0]["border"]= largest_string - panel.GetTextExtent(item[0]["name"])[0] + 5
                object = item[1](panel, **item[0])
                if ToolTips:
                    try:
                        object.SetToolTipString(Documentation[item[0]["name"]])
                    except KeyError:
                        print "No documentation for %s." % repr(item[0]["name"])
                object.end_init()
                
                if (type(object) == widgets.OpenButton or type(object) == widgets.SaveRunButtons
                    or type(object) == widgets.PlotPanel or type(object) == widgets.Text
                    or type(object) == widgets.ExpandableOnOff):
                    panel.Sizer.Add(object, 0, flag=wx.ALL|wx.EXPAND, border=5)
                elif (type(object) == widgets.Html):
                    panel.Sizer.Add(object, 1, wx.EXPAND|wx.ALL, border=5)
                else:
                    panel.Sizer.Add(object, 0, flag=wx.ALL|wx.EXPAND, border=5)
                # Check if the object has an event handler
                if item[-1]:
                    object.Bind(item[-1][0], item[-1][1])
                out[item[0]["name"]] = object
            out[first].SetFocusFromKbd() # because of the SpinCtrl feature/BUG!!
            panel.SetSizer(panel.Sizer)
            return panel, out

class Main(wx.Frame):
    """Main Frame

    WARNING, the documentation is mostly oudated.
    
    The main frame contains most of the functionality needed for saving,
    opening, running and plotting. Most of the design is also created here.

    Some info about the different dictionaries and Options.
    The self.Options dictionary will after the Main.__init__ method have
    completed, contain all options, such as albedo, wavelength etc.
    The Other dictionary contains everything that don't fit inside
    the Options dictionary, like input_file, out_file and classes
    that probably don't really need to be stored in a dicitonary, like run and Plot.
    The self.Other and self.Options dicitonaries are very important for the GUI
    The tuples plot, general and create_list contains information about how
    the different options and various gui elements will be added. The "format"
    of the tuples with information is a little bit caothic to understand,
    especially for text editors without higlightning of bracket pairs(like mine!).
    ((Page_Name), ({This dictionary will contain different arguments to the
    specific widget}, this_will_be_the_widget_type,
    (wx.OPTIONAL_EVT_TO_BIND_IN_THE_WIDGET, self.EVENT_HANDLER)))
    Every page starts with a tuple which contains the name of the page, and then
    none or more other tuples containing the different other options.
    The options in plot and general will be saved in self.Other
    The options in create_list will be saved in self.Options
    At the moment the page, the parent of every widget/option is not stored anywhere.
    To add a new option simply copy for instance the line concerning albedo.
    Change the needed arguments in the tuple and maybe the widget type.
    More kinds of widgets will be added later in the widgets.py file
    The behaviour of each widget can be seen by running widgets.py
    """
    global Options
    Options = {}    # self.Options contains all options
    global Other
    Other = {} # "old_inp" and "old_order" will contain the old order
                                           # of .INP files that have been read
    Saved = True # Start as saved, changes if anything is changed
    def __init__(self, quick_start=False):
        #-----------------------------------------------------------------------
        # Be careful with the brackets when editing the lists
        #plot = ( 
        #    ( ("Plot",),
        #        ({"name":"Plot", "info":"Plot", "ButtonLabel":"Plot"}, widgets.InfoButton, (wx.EVT_BUTTON, self.OnPlot)),
        #        ({"name":"out_file", "format":"*.OUT"}, widgets.OpenButton, ()),
        #        ({"name":"plot_choices"}, PlotChoices, ()),
        #        ),
        #        )
        #-----------------------------------------------------------------------
        general = (
            ( ("General",),
              ({"name":"Input File", "without_clear_button":True}, widgets.SaveRunButtons, (wx.EVT_BUTTON, self.OnSave)),
                # Event binding of the SaveButtons takes place later
              ({"name":"Output File","SaveButtonLabel":"Run", "format":".OUT", "SaveAsButtonLabel":"Save and Run", "without_clear_button":True}, widgets.SaveRunButtons, (wx.EVT_BUTTON, self.OnRun)),
              #({"name":"Plot_Out", "Other1":Other, "Options1":Options, "info":"Output File", "format":"*.*", "choices":["direct irradiance (red)", "diffuse downward irradiance (green)", "upward irradiance (blue)", "radiances"], "without_clear_button":True}, widgets.PlotPanel, (),),
              ({"name":"How To", "filename":pr("resources/how_to.html")}, widgets.Html, ()),
              ({"name":"new", "info":"New Input File", "ButtonLabel":"New", "without_clear_button":True}, widgets.InfoButton, (wx.EVT_BUTTON, self.OnNew)),
                ),
                )
        #-----------------------------------------------------------------------
        create_list = create.create_list # Edit create.py when adding new options
        #-----------------------------------------------------------------------
        wx.Frame.__init__(self, None, title=window_title, size=window_size)

        self.error_txt = "" # most errors should be collected here before they are displayed,
                                # it is better with only one long error message than ten small ones

        global tmp_panel
        tmp_panel = wx.Panel(self)
        tmp_panel.Show(False)
        
        self.Start = Start(self)
        self.Edit = NoteBook(self) # Edit contains all the pages
        if not quick_start:
            self.Edit.AddPages(general, Other)    # Create the General tab

            create_list.append((("Plot",),({"name":"Plot_Out", "info":"Output File", "format":"*.*", "choices":["direct irradiance (red)", "diffuse downward irradiance (green)", "upward irradiance (blue)", "radiances"], "without_clear_button":True}, PlotPanel, (),),),)
            self.Edit.AddPages(create_list, Options) # Create all the other tabs
            
        self.Edit.Show(False)# Only show Edit after either OnNew or OnOpenExistingINP have been run
        
        self.HorSizer = wx.BoxSizer(wx.VERTICAL)
        self.HorSizer.Add(self.Start, 1, wx.EXPAND|wx.BOTTOM|wx.RIGHT|wx.LEFT, border=5)
        
        self.SetSizer(self.HorSizer)

        self.Center(True)
        self.SetSizeHints(*window_size)

        global DocWindow
        if DocWindow:
            DocWindow = widgets.DocWindow = widgets.DocumentationWindow(self, doc_dir=pr("resources/html_doc"))

        self.Show(True)

        self.Clear() # Because of the intspinner
        
        # Bind some special events
        self.Bind(wx.EVT_BUTTON, self.OnNewInp, self.Start.New)
        self.Bind(wx.EVT_BUTTON, self.OnOpenExistingINP, self.Start.Open)
        self.Bind(wx.EVT_BUTTON, self.OnExample, self.Start.TryButton)
        self.Bind(wx.EVT_CLOSE, self.OnExit, self)

    def Has_Change(self, event):
        """Mark that an option has changed and the document needs to be saved"""
        object = event.GetEventObject().GetParent()
        if object:
            object.Has_Changed = True
            self.Saved = False
        event.Skip()
    def OpenFile(self, fname):
        self.SetOptions(self.ReadFile(fname))
        Other["Input File"].SetValue(fname)
    def OnNew(self, event):
        if self.GuardianAngel():
            self.Clear()
            self.HorSizer.Detach(self.Edit) # Remove the start panel
            self.Edit.Show(False)
        
            self.HorSizer.Insert(0, self.Start, 1, wx.EXPAND)   # Insert the editing panel
            self.Start.Show(True) # Make it visible
        
            self.HorSizer.Layout()
            self.Refresh()
    def Clear(self):
        for option in Options:
            Options[option].Clear()
        for special in Other:
            Other[special].Clear()
        widgets.Saved = True    
    def OnNewInp(self, event):
        """When the New Button is pressed, self.Main.Start.New"""
        tmp = FileDialog(format="*.INP", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
        if tmp:
            Other["Input File"].SetValue(tmp)
            self.Start_End()
    def OnOpenExistingINP(self, event):
        """When open is pressed on the start page"""
        tmp = FileDialog(format="*.INP")
        if tmp:
            self.OpenFile(tmp)
            self.Start_End()
    def OnExample(self, event):
        example = self.Start.Examples.GetCurrentPage()
        self.OpenFile(example.example_path)
        self.Start_End()
    def Start_End(self):
        """Removes the start panel and inserts the Edit panel"""
        self.HorSizer.Detach(self.Start) # Remove the start panel
        self.Start.Show(False)                      # and the destroy it
        
        self.HorSizer.Insert(0, self.Edit, 1, wx.ALL|wx.EXPAND, border=5)   # Insert the editing panel
        # might be changed with wx.Sizer.Prepend
        self.Edit.Show(True) # Make it visible
        
        self.HorSizer.Layout()
        self.Refresh()      # To prevent some ugly redraw errors

    def OnExit(self, event):
        """This event is triggered on exit
        
        It might be bound later to at least three events.
        wx.EVT_CLOSE  When the fram is closed
        wx.EVT_BUTTON When the quit button is pressed(quit button will be added later)
        wx.EVT_MENU  When quit is chosen in the Menu
        
        The event should probably have som check if anything have been edited
        since last save and only prompt the user for a save dialog then
        """
        if widgets.Saved:
            self.Destroy()
            return
        else:
            if self.GuardianAngel():
                self.Destroy()
    def GuardianAngel(self):
        """ Protect the user from losing unsaved work."""
        Dialog = wx.MessageDialog(None, "Save Input file before exit?", "Save",
                        wx.YES_NO|wx.CANCEL|wx.YES_DEFAULT)
        answer = Dialog.ShowModal()
        if answer == wx.ID_YES:
            save_file = Other["Input File"].GetValue()
            if not save_file:
                save_file = FileDialog(format=".INP", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
                Other["Input File"].SetValue(save_file)
            self.Save(save_file)
            widgets.Saved = True
            Dialog.Destroy()
        elif answer == wx.ID_CANCEL:
            Dialog.Destroy()
            return False
        return True
    def OnSave(self, event):
        tmp = event.GetEventObject().GetParent().GetValue()
        if not tmp:
            tmp = FileDialog(".INP", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
            if not tmp: return
            event.GetEventObject().GetParent().SetValue(tmp)
        self.Save(tmp)
        widgets.Saved = True
    def OnRun(self, event):
        out_file = Other["Output File"].GetValue()
        if not out_file:
            out_file= FileDialog(".OUT", flags=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
            if not out_file: return
            Other["Output File"].SetValue(out_file)
        # CHANGE! Use tmpfile
        tmp_file = os.path.abspath(".tmp_UVSPEC.INP")
        tmp_save = self.Saved
        self.Save(tmp_file) # Save a tempoary INP file with the latest modifications
        self.Saved = tmp_save # Set it to the previous value, since self.Saved sets it to True

        win = widgets.RunUvspec(self, tmp_file, out_file)
        Options["Plot_Out"].SetValue(out_file)
    def env_check(self):
        """Check if the environment variable and uvspec exists
        
        If the environment varaible "LIBRADTRAN_DATA_FILES" don't exist and uvspec 
        can't be found in the path the run button will be disabled.
        """
        print "env_check"
        if not os.getenv("LIBRADTRAN_DATA_FILES"):
            warning = wx.MessageDialog(None, 
            """The Environmentvariabel LIBRADTRAN_DATA_FILES was not found!
            You can still edit .INP files and plot .OUT files, but you can not run UVSPEC! It is strongly recomended to set the environment variabel and restart the GUI!""",
            "Warning!", # Title
            wx.OK)      # Style
            warning.Center(True)
            warning.ShowModal()
            # Disable run Button
            Other["Output File"].Enable(False)
        try:
            subprocess.Popen(("uvspec", "-v"), stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            # remove all output the command
        except OSError:
            warning = wx.MessageDialog(self, 
            """Could not find UVSPEC in your PATH!
            You can still edit .INP files and plot .OUT files, but you can not run UVSPEC! It is strongly recomended to change your PATH and restart the GUI!""",
            "Warning!", # Title
            wx.OK)      # Style
            warning.Center(True)
            warning.ShowModal()
            # Disable run Button
            Other["Output File"].Enable(False)
    def Save(self, fname):
        data = "" + "#Generated by libRadtran GUI(%s)\n" % version
        for option in Options:
            if Options[option].Has_Changed:
                values = Options[option].GetValue()
                if values is None: continue
                if isinstance(values, basestring): values = (values,)
                data += ("%s " % option) + " ".join(values) + "\n"
        try:
            print "saving input file to '%s'." % (fname)
            print "'%s'" % data
            f = open(fname, "w")
            f.write(data.strip())
            f.close()
        except IOError:
            msg = "Could not save input file to '%s'." % (fname)
            print msg
            ErrorMessage(msg)
                
    def ReadFile(self, path, includes=[]):
        try:
            f = open(path, "r")
            data = f.readlines()
            f.close()
        except IOError:
            msg = "Could not open %s." % (repr(path))
            self.error_txt += msg
            print " "*len(includes) + msg
            return

        data = [line.split() for line in data]

        # Remove lines with comments and merge continous lines
        line = 0
        while line < len(data):
            if not data[line]: data.pop(line); continue # skip empty lines
            if data[line][0].startswith("#"): data.pop(line); continue # skip comments
            elif [True for word in data[line] if (word.find("#") != -1)]: # remove comments from the line
                tmp = []
                for word in data[line]:
                    pos = word.find("#")
                    if pos != -1:
                        if word != "#":
                            tmp.append(word[:pos])
                        break
                    else:
                        tmp.append(word)
                data[line] = tmp
            elif data[line][-1].endswith("\\"): # continous line
                data[line][-1] = data[pos][-1][:-1] # remove the \
                if data[line][-1] == "": data[line].pop() # if the \ was preceded and continued by whitespace
                data[line].extend(data[line + 1])
            else:
                line += 1

        # Get the includes and include them
        buff = 0
        this_includes = []
        for line in xrange(len(data)):
            if (data[line + buff][0] == "include" and len(data[line + buff]) == 2):
                this_includes.append(data.pop(line + buff)[1])
                buff -= 1

        for include_path in this_includes:
            if not os.path.exists(path):
                msg = "Include %s in %s does not exist.\n" % (include_path, path)
                self.error_txt += msg
                print " " * len(includes) + msg
            elif (this_includes + includes).count(include_path) != 1: # if the file has been included before
                msg = "Include %s included more than one time. One time in %s. Please fix this " % (include_path, path)
                self.error_txt += msg
                print " " * len(includes) + msg
            else:
                include_data = self.ReadFile(include_path, includes + this_includes)
                if include_data: # the includ file might contains errors and return None
                    data.extend(include_data)
        
        if self.error_txt and not includes: # only Display error message in the most outer scope
            ErrorMessage(self.error_txt)
            self.error_txt = ""

        return data

    def SetOptions(self, data):
        self.error_txt = ""
        for line in data:
            try:
                error = Options[line[0]].SetValue(line[1:])
                if error: self.error_txt += error + "\n"
            except KeyError:
                error = "Unknown options '%s'.\n" % line[0]
                self.error_txt += error
                print error
        if self.error_txt:
            ErrorMessage(self.error_txt)

def main(GUI_path="", examples_path=".."):
    global libRadtran_GUI_path
    libRadtran_GUI_path = GUI_path
    global libRadtran_examples_path
    libRadtran_examples_path = examples_path

    create.create_example_list()
    
    app = wx.App(redirect=redirect, filename=redirect_file)
    print time.strftime("%Y-%m-%d-%H-%M-%S")
    splash = SplashScreen()
    frame = Main(quick_start=False)
    app.SetTopWindow(frame)
    if splash: splash.Close() # close directly after all GUI elements are initialized
    frame.env_check() # have to be called after the Splashscreen is closed
                               # because the MessageDialog and Splashscreen 
                               # will block each other
    """
    #useful when testing the GUI
    Other["Input File"].SetValue("/Users/jonas/tmp.inp")
    data = frame.ReadFile(Other["Input File"].GetValue())
    frame.Attribute(data)
    frame.Start_End()
    Other["Plot_Out"].SetValue("/Users/jonas/tmp.out")
    """
    app.MainLoop()

if __name__ == "__main__":
    main()
