/************************************************************************/
/* mystic.h                                                             */
/*                                                                      */
/* MYSTIC - Monte Carlo code for the physically correct tracing of      */
/*          photons in cloudy atmospheres.                              */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*         DLR, bernhard.mayer@dlr.de                                   */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 2000 Bernhard Mayer                                    */
/************************************************************************/

#ifndef __mystic_h
#define __mystic_h

#if defined (__cplusplus)
extern "C" {
#endif

/* for debugging. This mode writes the randomstatus file, with */
/* help of it one can start shortly before a buggy photon      */
/* instead of having to simulate many good photons before      */
/* reaching the problematic one                                */
#undef WRITERANDOMSTATUS

#undef MUCHOUT

#include <stdio.h>  /* required for FILENAME_MAX */
#include "phasetable.h"

#define NOT_A_NUMBER              0.0/0.0

/* radiation source */
#define MCSRC_NONE                0
#define MCSRC_SOLAR               1
#define MCSRC_THERMAL_SURFACE     2
#define MCSRC_THERMAL_ATMOSPHERE  3
#define MCSRC_THERMAL_BACKWARD    4/*TZ bt*/
#define MCSRC_LIDAR               5
#define MCSRC_DETECTOR            6

/* albedo method */
#define MCALB_NONE                0
#define MCALB_LAM                 1
#define MCALB_LAM2D               2
#define MCALB_RPV                 3
#define MCALB_RPV2D               4
#define MCALB_COXANDMUNK          5
#define MCALB_AMBRALS             6
#define MCALB_AMBRALS2D           7
#define MCALB_TSANG               8


/* boundary conditions */
#define MCBCOND_NONE              0
#define MCBCOND_PERIODIC          1
#define MCBCOND_MIRROR            2
#define MCBCOND_ABSORB            3

/* calculate 3D absorption field? */
#define MCFORWARD_ABS_NONE        0    /* no absorption                     */
#define MCFORWARD_ABS_ABSORPTION  1    /* calculate absorption              */
#define MCFORWARD_ABS_EMISSION    2    /* calculate emission                */
#define MCFORWARD_ABS_HEATING     3    /* calculate (absorption + emission) */
#define MCFORWARD_ABS_ACTINIC     4    /* calculate actinic flux            */

#define MCABS_UNIT_W_PER_SQUARE_M_AND_DZ  0  /* default absorbed energy per box */
#define MCABS_UNIT_W_PER_CUBIC_M          1  /* normalised with height          */
#define MCABS_UNIT_K_PER_DAY              2  /* converted to K / day            */

#define MCBACKWARD_NONE           0 /* forward calculation           */
#define MCBACKWARD_RADIANCE       1 /* backward radiance             */
#define MCBACKWARD_EDIR           2 /* direct irradiance             */
#define MCBACKWARD_EDN            3 /* downward diffuse irradiance   */
#define MCBACKWARD_EUP            4 /* upward diffuse irradiance     */ 
#define MCBACKWARD_FDIR           5 /* direct irradiance             */
#define MCBACKWARD_FDN            6 /* downward diffuse irradiance   */
#define MCBACKWARD_FUP            7 /* upward diffuse irradiance     */ 
#define MCBACKWARD_ABS            8 /* absorption                    */
#define MCBACKWARD_ACT            9 /* actinic flux                  */
#define MCBACKWARD_EMIS           10 /* emission                      */
#define MCBACKWARD_HEAT           11 /* heating rate                  */

/* flag what is defined in the wcloud3D.dat, 1,2,3 */
#define CLD_OPTPROP               1 /* explicit optical properties: ext, asy, ssa */
#define CLD_EXTREFF               2 /* extinction + effective radius              */
#define CLD_LWCREFF               3 /* liquid water content + effective radius    */


/* azimuth conventions */
#define MCAZIMUTH_NEW             0
#define MCAZIMUTH_OLD             1

/* small number for floating point comparisons */
#define MC_EPSILON                1E-6

/* scatter types */
#define MCPROF_SCATTYP3D          2 /* this should only include 3D arrays, these must be the first ones starting from 0 */
#define MCPROF_SCATTYPES          4 /* this should not include _NONE, _VIRT, or _TOT */
#define MCPROF_VIRT      MCPROF_SCATTYPES+1
#define MCPROF_NONE               -1
#define MCPROF_TOT                0
/* the following are 3D physical scatter types */
#define MCPROF_WC                 1
#define MCPROF_IC                 2
/* the following are 1D physical scatter types */
#define MCPROF_MOL                3
#define MCPROF_AER                4

/* status of photon */
#define MCSTATUS_DEFAULT          0
#define MCSTATUS_TRAVEL           1
#define MCSTATUS_SCATTER          2 
#define MCSTATUS_ABSORB           3 
#define MCSTATUS_SURFACE          4 
#define MCSTATUS_OUTOFDOMAIN      5 
#define MCSTATUS_BOUNDARY_LOWER   6 
#define MCSTATUS_BOUNDARY_UPPER   7 
#define MCSTATUS_HIT_LIDAR        8 
#define MCSTATUS_HIT_VIS_FOD_CONE 9 
#define MCSTATUS_HIT_DET_PLANE    10
#define MCSTATUS_VIRTUALSCATTER   11
#define MCSTATUS_VISQIDD_ACTION   12
#define MCSTATUS_RISQIDD_ACTION   13
#define MCSTATUS_INCREMENT        14
#define MCSTATUS_CLONE            15
#define MCSTATUS_SPLIT            16
#define MCSTATUS_PURGE            17

/* mode of Scattering, e.g. Delta Scaling */
#define MCSC_MODE_NORMAL          0
#define MCSC_MODE_DELTA_SCALE     1

/* mode of Real Importance Sampling */
#define MCRIS_MODE_NORMAL         0
#define MCRIS_MODE_MAS            1

/* mode of Virtual Importance Sampling */
#define MCVIS_MODE_NORMAL         0
#define MCVIS_MODE_QIDD           1

/* mode of Detector Directional Importance Sampling */
#define MCDDIS_NONE               0
#define MCDDIS_UPF                1
#define MCDDIS_UDA                2

/* version of jacobian */
#define MC_JAC_VAR_BETA           0
#define MC_JAC_VAR_REFF           1
#define MC_JAC_VAR_ABS            1 /* 1 on purpose! */

/* number of phase matrix elements */
#define NPHAMAT                   6

typedef enum {NONE, UPPER, LOWER} BOUNDARY;

typedef struct {
  double a;
  double b;
  double c;
  double d;
  double upper;
} surface;

/******************************************************************************************/
/* Description of profile structures:                                                     */
/* =================================  						          */
/*  											  */
/* The atmospheric profiles have become a much more compact and flexible structure. 	  */
/*  										          */
/* A typical call to a profile data point will look like this: 			          */
/*  											  */
/* atmos->ksca   [p->SC_mode][p->RIS_mode] ->prof [ist][p->kc] 			          */
/* atmos->ksca3D [p->SC_mode][p->RIS_mode] ->prof [ist][p->kc][p->ic][p->jc] 		  */
/*  											  */
/* Here, ksca and ksca3D stand for the scattering rates in 1D and 3D, resp., 		  */
/* p->SC_mode is the mode of scattering, e.g. NORMAL or DELTA_SCALE, 			  */
/* p->RIS_mode is the mode of real importance sampling, e.g. NORMAL or 		          */
/*                                 enhancement of molecular scattering, 		  */
/* prof[ist] is the profile structure of particle type ist, 				  */
/* ist is the particle type, e.g. WC, IC, MOL, AER, TOT (=total) 			  */
/*  											  */
/* A new variable is kext: 								  */
/*  											  */
/* atmos->kext   [p->SC_mode][p->RIS_mode][p->VIS_mode] ->prof [ist][p->kc] 		  */
/* atmos->kext3D [p->SC_mode][p->RIS_mode][p->VIS_mode] ->prof [ist][p->kc][p->ic][p->jc] */
/*  											  */
/* Here, only ist=TOT exists. 							          */
/* p->VIS_mode is the mode of virtual importance sample, e.g. NORMAL or 		  */
/*                                    enhancement inverse to detector distance 	          */
/*  											  */
/* kext... is the interaction rate that is used to reduce tau in travel_tau 		  */
/*  											  */
/* ksca...[RIS_mode != NORMAL] is the interaction rate used to choose scattering 	  */
/*        ( if ksca < kext, virtual scatter can happen. These do local and 		  */
/*          directional estimates, but no change of photon direction       ) 		  */
/*  											  */
/* ksca...[p->RIS_mode == NORMAL] is the true scatter rate 				  */
/*        ( if ksca[NORMAL] < ksca[RIS_mode] the photon weight changes: 		  */
/*          the ratio times exponential of the difference times pathlength )  	          */
/*  											  */
/******************************************************************************************/

typedef struct {
  int ntypes;
  int n;
  double **prof;
} profile;

 
/* profile3D is required to store the profiles of  */
/* absorption and scattering coefficients; float   */
/* should be accurate enough                       */

typedef struct {
  int ntypes;
  int Nz;
  int Nx;
  int Ny;
  int *threed;     /* flag if a layer is 3D */
  int nthreed;     /* number of 3D layers   */
  float ****prof;
  int *tocalloc;
} profile3D;


/* ddprofile3D is required to store the profile of */
/* the absorbed energy; as all output quantities,  */
/* this parameter is also stored in double rather  */
/* than float                                      */

typedef struct {
  int Nz;
  int Nx;
  int Ny;
  int *threed;
  int nthreed;
  /* double ***aer; */
  /* double ***ozo; */
  /* double ***ice; */
  double ***cld;
  double ***tot;
} ddprofile3D;


typedef struct {
  double mol;
  double aer;
  double cld;
  double ice;
  double tot;
} optical_depth;


typedef struct {
  int    **ndir;    /* number of photons     */
  int    **ndn;
  int    **nup;

  double **edir;     /* irradiance           */
  double **edn; 
  double **eup;  

  double **fdir;     /* actinic flux         */
  double **fdn; 
  double **fup;  

  /* CE included additional dimension for all radiances to account for polarisation */
  double ****raddir;  /* radiance             */
  double ****raddif;
  double ****radesc;

  /* introduced for I3RC case7: radiance binned in radial and pathlength bins */
  double ****radpat; /* cone sampling  */
  double ****radpes; /* local estimate */

  /* standard deviations */
  int std_alloc;     /* flag if allocated */
  double **edir2;    /* irradiance        */
  double **edn2; 
  double **eup2;  
  
  double **fdir2;    /* actinic flux      */
  double **fdn2; 
  double **fup2;  

  double ****raddir2; /* radiance          */
  double ****raddif2;
  double ****radesc2;
  double ****radpat2; 
  double ****radpes2; 

  int ****nraddir;    /* number of photons */
  int ****nraddif;    /* (radiance)        */
  int ****nradesc;
  int ****nradpat;
  int ****nradpes;
} radiation_field;

  
typedef struct {
  double k;
  double rho0;
  double theta;
  double scale;
  double sigma;
  double t1;
  double t2;
} rpv_struct;


typedef struct {
  double iso;
  double vol;
  double geo;
  int isCaM;
} ambrals_struct;


typedef struct {
  int method;
  char filename[FILENAME_MAX];

  double **albedo2D;
  double albedo;
  double delX;
  double delY;
  int Nx;
  int Ny;
  double Wsurf;

  int reflectalways;

  char rpv_filename [FILENAME_MAX];
  rpv_struct *rpv;   /* rpv structure                              */
  int *rpv_count;    /* number of occurrence of each RPV index     */
  char **rpv_index;  /* xy-field of RPV indices                    */

  /* Cox and Munk */
  float u10;
  float pcl;
  float xsal;
  float uphi;
  int   solar_wind;

  char ambrals_filename [FILENAME_MAX];
  ambrals_struct **ambrals2D;  /* xy-field of Ambrals BRDF structure */
  ambrals_struct *ambrals;     /* homogeneous Ambrals BRDF */

} albedo_struct;


typedef struct photon_struct *photon_sub_struct;
                    /* this trick is needed in order to define a */
		    /* photon_struct within locest_struct within */
                    /* a photon_struct                           */


typedef struct {
  double dx[3];
  int hit[3];
  double cotheta; 
} direction;


typedef struct {
  direction dir;           /* direction                                  */

  double cosalpha;         /* cosine of detector surface angle           */
  double pdir;             /* probability                                */
  double pdir_iso;         /* probability for isoenergetic scattering    */
  double *pdir_sct;
  double *pdir_jac;

  double dist;             /* distance to hit point (detector/beam end)  */
  double distinv;          /* inverse distance to hit point              */

  double r_det;
  double z_det;
  double x_cc[3];
  double t_det;

  double x_hit[3];
  double weight_hit;

  int in_cone;
  int will_hit_cone;
  int behind_detector;
  int will_hit_det_plane;
  int lidar_outside_grid;

  double hit_det_plane_step; /* distance till photon hits detector plane      */
  double vis_fod_step;       /* distance till something happens for VIS-FOD */
  double vis_fod_step2;      /* distance till something happens for VIS-FOD */
  double vis_fod_kext;       /* interaction rate inside VIS-FOD cone        */
			     
  photon_sub_struct hitpoint; /* the detector hit point to be hit in         */
                              /* local estimator mode                        */

} locest_struct;

/* structure to store the number of local estimates of certain weight "add" */
struct tnode_LE {
  int add;	       /* approximate weight of bin */
  int nevents;         /* sum of events within bin  */
  struct tnode_LE *left;  /* left child                */
  struct tnode_LE *right; /* right child               */
};

typedef struct {
  int ntypes;
  int norders;
  int nz;
  int nt;
  int np;
  int no;
  int na;
  double ******radloc;     /* radiance for lidar                      */
  double ******radloc2;    /* standard deviation                      */
  double *******radloc_t;  /* radiance for lidar                      */
  double *******radloc_t2; /* standard deviation                      */
  double ******jacobian_t;
  double ******jacobian_t2;
  double *****jacobian_t2act;
  int *actis;
  double **photoncounter;  /* number of photons that are simulated    */
  struct tnode_LE ****dtree; /* local estimator size count tree    */
} lidar_result_field;

typedef struct {
  int Nx;
  int Ny;
  int Nz;
  double ****jacobian_t;
  double ****jacobian_t2;
  double **jacobian_tact;
//  double **photoncounter;  /* number of photons that are simulated    */
//  struct tnode_LE ****dtree; /* local estimator size count tree    */
} jacobian_result_field;

  typedef struct {
    int Nv; 
    int Nx; 
    int Ny; 
    int Np;
    double ****rad_t;        /* spectral radiation field [Nv,Nx,Ny,Np] */
  } radiation_field_t;
  

  typedef struct {
    radiation_field *surf;
    radiation_field **alt;
    lidar_result_field *lidar;
    jacobian_result_field *jacobian;
    radiation_field_t *surf_t;
    radiation_field_t **rad_t;   
    
    double ***back;
    double ***back2;     /* BCA: Maybe a comment on what is back2? */
    double **backemis;
    
    ddprofile3D *absorption3D;    /* average            */
    ddprofile3D *absorption3D2;   /* standard deviation */
  } result_struct;


/* Structure to store a direction and opening angle */
/* for radiance calculations; convention:           */ 
/*    phi=0,   South                                */
/*    phi=90,  West                                 */
/*    phi=180, North                                */
/*    phi=270, East                                 */
/*                                                  */
/*    theta > 0 upward                              */
/*    theta < 0 downward                            */

typedef struct {
  double theta;         /* polar angle                      */
  double phi;           /* azimuth, internal                */
  double externalphi;   /* azimuth, as defined by the user  */
  double alpha; 
  double cosalpha; 
  double omega; 
  direction dir;
} radang;

typedef struct {
  double theta;         /* polar angle                      */
  double phi;           /* azimuth, internal                */
  double externalphi;   /* azimuth, as defined by the user  */
  double alpha; 
  double cosalpha; 
  double omega; 
  direction dir;

  double x[3];          /* position                           */
  double radius;        /* radial size of laser/detector area */
  double area;          /* laser/detector area                */
  int divergence;

} lidar_laser;

typedef struct {
  double theta;         /* polar angle                      */
  double phi;           /* azimuth, internal                */
  double externalphi;   /* azimuth, as defined by the user  */
  double alpha; 
  double *cosalpha; 
  double *omega;
  double *azphi;
  direction dir;
  double dx_perp1[3];
  double dx_perp2[3];

  double x[3];          /* position                           */
  double radius;        /* radial size of laser/detector area */
  double area;          /* laser/detector area                */

  double s_det;
  double c_det;
  double t_det;
  double z_det;
  double x_cc[3];

} lidar_detector;


/* structure to store the sum of photon weights for an individual photon */
struct tnode {
  double *loc;	  /* pointer to storage    */
  double weight;  /* sum of photon weights */
  struct tnode *left;       /* left child  */
  struct tnode *right;      /* right child */
};

/* structure to store the sum of photon weights for an individual photon */
struct anode {
  double *loc;	  /* pointer to storage    */
  double *array;  /* sum of photon weights */
  struct anode *left;       /* left child  */
  struct anode *right;      /* right child */
};


/* linked list element to store a photon path */
struct photon_path {
  double x[3];
  struct photon_path *next;
};


typedef struct photon_struct {
  /* current status of photon */
  double x[3];               /* photon location                              */
  direction dir;             /* photon direction                             */
  direction dir0;            /* initial photon direction                     */
  direction dir_ort;         /* orthogonal direction to scattering plane     */

  double lest_phi; 	/* Local estimate direction in regard to center of sun */
  double lest_alpha; 	/* Local estimate direction in regard to center of sun */
  
  int ic;                    /* box index, x                                 */
  int jc;                    /* box index, y                                 */
  int kc;                    /* box index, z                                 */

  int photon_status;         /* flag: current status of photon               */
  int reallyhit[3];          /* flag: crossed box boundary                   */
  int muchoutcounter;

  int SC_mode;               /* actual mode for Scattering                   */
  int DDIS_SC_mode;          /* actual mode for DDIS Scattering              */
  int RIS_mode;              /* actual mode for Real Importance Sampling     */
  int VIS_mode;              /* actual mode for Virtual Importance Sampling  */

  /* counters */
  int direct;                /* direct or diffuse                            */

  int scattercounter;        /* number of scattering + reflection events     */ 
  int photoncounter;         /* number of current photon                     */
  int rayleighcounter;       /* count number of rayleigh scatters            */
  int isclone;
  int clonescattercounter;
  int escapescattercounter;
  int spikewarningcounter;

  double pathlength;         /* photon pathlength                            */
  optical_depth tauabs;      /* absortion optical depth along the path       */
  double tauris;
  double p_norm;
  double *pathlength_per_layer; /* photon pathlength per layer               */       

  /* weights */
  double weight;              /* general weight, contains what is not below   */
  double *stokes0;            /* initial stokes vector for polarization       */
  double *stokes;             /* stokes vector for polarization               */
  double **phamat;            /* phase matrix of photon                       */
  double q_isoene;            /* weight for isoenergetic scattering           */
  double ****q_jacobian;      /* weight for jacobian                          */
  double *r_jacobian;         /* pathlength for jacobian                      */
  int nlyr;                   /* number of atmosph. layers, needed by jacob.  */
  /* CE replaced the following by pathlengths in each layer */
  /* double *dtauabs_spectral;    weight vector for spectral absorption       */ 
  int nlambda;                /* number of wavelengths included in q_spectral */
  double *q_spectral;         /* weight for spectral calculation              */
  double *q2_spectral;
  double special_weight;

  /* stuff for local estimator / escape radiances */
  double *pdir;              /* probability for direction radang[]           */
  locest_struct lest;        /* variables needed for local estimator         */

  /* honestly, these should not be in this structure */
  int ipa;                   /* independent pixel approximation              */
  double maxpathlength;      /* evtl only allow this as maximal pathlength   */

  /* miscellaneous */
  struct tnode *wtree;       /* photon weight tree                           */
  struct anode *wtree2;       /* photon weight tree                          */
  struct photon_path *path;  /* photon path linked list                      */

  double vis_beta;
  double risqidd_beta;

  photon_sub_struct parent_photon;

  double tipa_start_tau;

} photon_struct;

typedef struct {
  int user_defined_btemp;      /* flag if user defined surface_temperature  */
  float btemp;                 /* constant surface temperature              */

  int surf2D; /* 2D temperature field */
  double **temp2D;
  double **plkavg2D;
  double delX;
  double delY;
  int Nx;
  int Ny;
} surftemp_struct;

typedef struct {
  int elev2D;

  surface **surf;
  double surfmax;
  double delX;
  double delY;
  int Nx;
  int Ny;

} elevation_struct;


typedef struct {
  double delX;
  double delY;
  double *X;
  double *Y;
  float  *Z;   /* all altitudes are floats to avoid roundoff problems */
  double *cos_lat;
  double *sin2_lat;
  double *tan2_lat;
  int Nx;
  int Ny;
  int Nz;
  double xmax;
  double ymax;
  int *threed;
  int nthreed;
  int nsca;     /* number of scatter modes, e.g. delta-scaling */
  int nscaRIS;  /* number of RIS scatter modes */
  int nscaVIS;  /* number of VIS scatter modes */
  double *ddis_eps;
  double r_earth;

  profile ***ksca;
  profile *kabs;
  profile ****kext;
  profile *g1;
  profile *g2;
  profile *ff;
  profile *reff;
  profile *Bplanck;
 
  double **kabs_spectral; /* spectral absorption coefficients */ 
  double **ksca_spectral;
  int nlambda_abs;
  float *lambda;
    
  profile3D ***ksca3D;
  profile3D *kabs3D;
  profile3D ****kext3D;
  int ***spiky_box;

  profile3D *g1cld3D;
  profile3D *g2cld3D;
  profile3D *ffcld3D;
  profile3D *reff3D;
  
  double rayleigh_depol; /* Rayleigh depolarisation */

  int nonHGcld;          /* Henyey-Greenstein or phase function table */
  int nonHGice;          /* Henyey-Greenstein or phase function table for ice TZ*/

  optical_depth tabstot;
  optical_depth tscatot;

  double Watm;
  double maxemis;

  int nztilt;
  float *tipa_transmission;
  float ***tipa_transmission3D;
  float *tipa_P_start_level;
  float ***tipa_P_start_level3D;

} atmosphere_struct;


typedef struct {
  int *sample;
  int nzout;
  radang *rad;
  direction rad_dir0; /* backup of rad.dir */
  double **umu2D;
  double **phi2D;
  int Nd;         /* number of angular directions */
  int Nr;         /* number of radii              */
  int Nt;         /* number of time intervals     */
  float dr;       /* radius interval              */
  float dt;       /* time interval                */
  double delX;
  double delY;
  int Nx;
  int Ny;
  int surface;
  int escape;
  int delta_scaling;
  int tipadir;
  
  double escape_eps_ddis_upf; /* DDISsing probability */
  int LidarLocEst;        /* local estimator for Lidar */
  int DetectorLocEst;        /* local estimator for Lidar */
  int ili;                /* active index of lidar detectors    */
  int reference_to_NN;        /* referencing to NN */

#if HAVE_VROOM
  int vroom;
  int LLE_RIS_MAS;        /* Real Importance Sampling for Molec. and Aerosol Scat. */
  int LLE_D_DIS;          /* Detector Directional Imp. Sampl.                      */
  int LLE_VIS_FOD;        /* Virt. Imp. Sampl. in FOV Of Detector                  */
  int LLE_VIS_QIDD;       /* Virtual IS according to Quadr. Inv. Dist. to Det.     */
  int LLE_sponti;         /* Spontaneous cloning                                   */
  int LLE_turnmax;        /* Turn off when converged                               */
  int LLE_moon;        /* Replace laser by moon or sun                           */
  int RIS_MS;
  int VIS_CS;
  int splitter;
  int ntupelLE;
  int startCP;
  int LEperCP;
  int use_p_norm;
  double n_split_max;
  double n_split_min;
  int LLE_jacobian;
  int abs_jacobian;
  int vroomreflectalways;

  int LLE_isoene;         /* output inelastic scatters */

  int LLE_RIS_MAS_start;
  int LLE_RIS_MAS_end;

  double LE_taucrit;      /* Iwabuchi critical tau                                 */

  int Nli;                /* number of lidar detectors          */
  int LLE_Nt;             /* number of time bins for lidar      */
  int LLE_No;             /* number of opening angles for lidar */
  int LLE_Na;             /* number of ring segments for lidar  */
  int LLE_islinear;       /* whether or not opening angles increase linearly */
  int LLE_polarisation;   /* polarisation of the laser beam     */
  int LLE_scatterout;
  double LLE_afac;        /* factor of opening angles for lidar */
  double LLE_pulse;       /* pulse length of the laser beam     */
  lidar_laser *laser;     /* emitter for local estimator        */
  lidar_detector *lidar;  /* detector for local estimator       */
  double *lidar_t;        /* time bins for differential loc.est.*/
  double *lidar_dt;       /* width of time bins                 */
  double ***jac_exp;      /* saver for results of jacobian exp  */
  int n_phase_max;
  pft **phase_max;

  double LLE_eps_ddis_upf;     /* DDISsing probability */
  double LLE_eps_ddis_uda;     /* DDISsing probability */
  double LLE_eps_fod_dis_phi;
  double LLE_taumax;
  double jac_eps;

  double visqidd_betamax;
  double visqidd_rmax;
  double visqidd_facs;
  double risqidd_betamax;
  double risqidd_facs;

  double MPdynDDIS;
  double split_max;
  double split_min;
#endif

  double LLE_stop_at_abserr;  /* stop simulation if loc.est. converged */
  double stop_at_relerr;     /* stop simulation if relative error small enough */
  double maxscatters;        /* evtl only allow this as maximum number of scatters */
  double minscatters;        /* evtl only allow this as maximum number of scatters */
  double maxpathlength;

  int passback3D;
  int std;        /* calculate standard deviation */

  int spectral_is;  /*spectral calculation with importance sampling */

  int spherical;
  int spherical3D;
  int refraction;

  int polarisation; 
  int nstokes;         /* Number of Stokes components */   

  int bcond;      /* boundary conditions */

  int backward; 
  int backward_is;
  int backward_js;

  int backward_islower;     /* sample pixels to be calculated; */
  int backward_jslower;     /* lower left - upper right        */
  int backward_isupper;
  int backward_jsupper;

  int backward_islower_org; /* same as above, to store original */
  int backward_jslower_org; /* data defined in the input file   */
  int backward_isupper_org; /* (they may be changed e.g. by     */
  int backward_jsupper_org; /*  forward2backward())             */

  int backward_writeallpixels;
  int backward_writeback;

  /*** forward tracing                             ***/
  double forward_sza;    /* solar zenith angle       */
  double forward_phi0;   /* solar azimuth angle      */
  double forward_vza;    /* viewing zenith angle     */
  double forward_phi;    /* viewing azimuth angle    */     

  float forward_zout;    /* output altitude          */
  float forward_zstart;  /* start altitude           */

  /*** backward tracing                            ***/
  double backward_sza;   /* solar zenith angle       */
  double backward_phi0;  /* solar azimuth angle      */
  double backward_vza;   /* viewing zenith angle     */
  double backward_phi;   /* viewing azimuth angle    */     

  float backward_zout;    /* output altitude         */
  float backward_zstart;  /* start altitude          */
  int sample_forward_kc;  /* forward sampling layer  */
  int sample_backward_kc; /* backward sampling layer */

  float* sample_backward_sunshape_p; /* distribution function of sunshape */
  pft    *sample_backward_sunshape_pft; /* pft structure */
  double* sample_backward_sunshape_F; /* cumulative distribution function of sunshape */
  double* sample_backward_sunshape_alpha; /* Angle alpha from center of sun in radians */
  int    sample_backward_sunshape_n_lines; /* Number of lines in sun shape file */
  int    sample_backward_sunshape; /* Switch */
  double sun_radius; /* Sun radius in degrees */

  /* actual angles */
  double sza;
  double phi0;
  double vza;
  double phi;

  float zstart;                 /* photon start position         */

  char backward_altstr[255];    /* store the altitude level for the output filename */

  int panorama;                 /*  panorama flag                */
  int pan_align;    /*  panorama flag                */
  int pan_align_to_sun;    /*  panorama flag                */
  int pan_align_to_mu;    /*  panorama flag                */
  int pan_no_pixel;        /*  panorama flag                */
  int pan_distr_photons_over_pixel; /*   panorama flag                */
  int pan_with_direct_rad; /*   panorama flag                */
  int pan_vr_sc0; 		/*   panorama flag                */
  double pan_umu_min;      /* panorama boundaries           */
  double pan_umu_max;      /* panorama boundaries           */
  double pan_theta_min;    /* panorama boundaries           */
  double pan_theta_max;    /* panorama boundaries           */
  double pan_phi_min;      /* panorama boundaries           */
  double pan_phi_max;      /* panorama boundaries           */
  double pan_dphi; 	/* panorama pixel width */
  double pan_dtheta; 	/* panorama pixel heigth */
  double pan_FOV;          /* Field of view of the actual panorama pixel in sr */
  double align_theta;
  double align_phi;

  int sensorposition;           /* user-defined sensor position  */
  double sensorposition_x[3];   /* only works in backward mode   */

  int sensordirection;          /* user-defined sensor-direction */
  double sensordirection_dx[3]; /* only works in backward mode   */

  int surfaceparallel;          /* sample surface-parallel       */

  double **surface_area;        /* surface area of sample pixels (2D elevation, surface-parallel, forward) */
  double **surface_area_std;    /* standard deviation                                                      */
  int **surface_area_counter;   /* counter for averaging                                                   */

} sample_struct;


/* 3d cloud structure; included in uvspec cld_out_struct; */
/* used by MYSTIC and IPA                                 */
typedef struct {
  int Nx;            /* number of grid points in x-direction */
  int Ny;            /* number of grid points in y-direction */
  int nlyr;          /* number of layers                     */
  
  double delX;       /* grid size in x-direction    [m] */
  double delY;       /* grid size in y-direction    [m] */
  
  float *zd;         /* level altitudes, 0 ... nlyr [m] */
                     /* all altitudes are float to      */
                     /* avoid roundoff problems         */

  float ***lwc;      /* liquid water content    [g/m3]   */
  float ***reff;     /* effective radius        [micron] */

  float ***ext;      /* extinction coefficient  [1/m]       */
  float ***ssa;      /* single scattering albedo            */
  float ***g1;       /* asymmetry parameter 1 for double-HG */
  float ***g2;       /* asymmetry parameter 2 for double-HG */
  float ***ff;       /* forward fraction for  double-HG     */
  float ***f;        /* delta scaling factor                */
  float ***dscale;   /* delta-scaling factor Tobias         */ 
  float reffmin;     /* minimum effective radius in cloud   */
  float reffmax;     /* maximum effective radius in cloud   */
  
  float ***tausol3d;  /* contains tau_wc or tau_ic at [nz][nx][ny], necessary for mc_tipa dir (ulrike) */
  int nztilt;         /* at how many levels tilting has to be performed = amount of "tilting-levels" */

  phase_function_table phase_cld;  /* complete single scattering properties: */
                                   /* tables of phase function and           */
                                   /* cumulative phase function              */            

  int cldproperties; /* as defined in the wcloud3D header                  */
  int nonHGcld;      /* flag: Henyey-Greenstein or explicit phase function */ 
  
  int *threed;       /* flag: 3D layer or 1D layer */
  int nthreed;       /* number of 3D layers        */

} cld3d_out_struct;


/* prototypes */
int mystic (int *nlyr,
	    float *dtair, float *dto3, 
	    float *dtaer, float *omaer, float *g1aer, float *g2aer, float *faer,
	    float *dtcld, float *omcld, float *g1cld, float *g2cld, float *fcld,
	    float *dscld, /*TZ ds*/
	    float *dtice, float *omice, float *g1ice, float *g2ice, float *fice,
	    float *dsice, /*TZ ds*/
            double **tau_molabs_user, float *lambda, int nlambda_abs,
            int rayleigh_crs, float rayleigh_depol_user, float *dens, float mixing_ratio_co2, 
            float *refind,
            float r_earth, 
	    float *rayleigh_depol,
	    cld3d_out_struct *wcld3d,
	    cld3d_out_struct *icld3d,
	    float *recld, float *reice,
	    float *temper, 
	    float *zprof,
	    float ***momaer, int *nmomaer,
	    int **nthetaaer, float ***thetaaer, double ***muaer, float ***phaseaer,
	    int nphamataer, int *spectral,
	    float *alb, float sza, float phi0, long int nphotons,
	    int *source, float *wvnmlo, float *wvnmhi, float *wavelength, 
	    float *zout, int *nzout,
	    float *rpv_rho0, float *rpv_k, float *rpv_theta,
	    float *rpv_scale, float *rpv_sigma, float *rpv_t1, float *rpv_t2,
	    float *ambrals_iso, float *ambrals_vol, float *ambrals_geo,
	    float *u10, float *pcl, float *xsal, float *uphi, int *solar_wind,
	    float *bpdf_u10, 
            int *ipa, int *absorption, int *loaddata,
	    sample_struct *sample,
	    elevation_struct *elev,
	    surftemp_struct *surftemp,
	    char *basename, 
	    char *umufilename,
	    char *sunshape_filename,
	    char *albedofilename,
	    char *rpvfilename, char **rpv_labels, int *rpv_nlabels,
	    char *ambralsfilename,
	    float *delta_scaling_mucut,/*TZ ds*/
	    float *truncate,
	    int *reflectalways,
	    int quiet,
	    float *rfldir, float *rfldn,  float *flup,
	    float *uavgso, float *uavgdn, float *uavgup,
	    float ***rfldir3d, float ***rfldn3d, float ***flup3d, 
	    float ***uavgso3d, float ***uavgdn3d, float ***uavgup3d, 
	    float ****radiance3d, float ***absback3d,
	    float ***abs3d, 	    
	    float ***rfldir3d_var, float ***rfldn3d_var, float ***flup3d_var, 
	    float ***uavgso3d_var, float ***uavgdn3d_var, float ***uavgup3d_var, 
	    float ****radiance3d_var, float ***absback3d_var,
	    float ***abs3d_var, 	    
	    int nxcld, int nycld, int nzcld,
	    char *datapath,
            char *randomstatusfilename,
	    int readrandomstatus,
	    int write_output_as_netcdf, 
	    int write_files, int visualize);
  
double sc_Rayleigh_mu();
double sc_Rayleigh_mu_depol (double depol);
double sc_Isotropic_mu();
double sc_Lambertian_mu();
double sc_HG_mu(double g);
double sc_HG2_mu(double g1, double g2, double ff);
double random_tau();
  

double rpv_brdf (double rho0, double k, double theta, 
		 double scale, double sigma, double t1, double t2,
		 double mu1, double mu2, double phi);

double rpv_albedo (double rho0, double k, double theta, double mu1);

int random_rpv_direction (double rho0, double k, double theta, double mu1,
                          double *mu2, double *phi);
  
photon_struct *calloc_photon (sample_struct *sample, int Nz, int source, int nlambda_abs);

void init_direction (double sinsza, double cossza, double sinphi,
                     double cosphi, 
                     direction *dir);

void gen_default_sunshape(float wvnmlo, float wvnmhi, int N, float *pd, double *alpha); 

int limb_dark_lest_dir (direction *dir_sun_center, sample_struct *sample,
			double *phi, double *alpha);

int random_ambrals_direction (double iso, double vol, double geo, 
			      double mu1, double *mu2, double *phi);

  void new_direction (double mu, double phi, direction *dir, double phi_inc);

void elev_coord (photon_struct *p, elevation_struct *elev, int *ie, int *je);

#if HAVE_LIDAR
int read_lidar (char *filename,  sample_struct *sample,
		int quiet);
#endif

#if defined (__cplusplus)
}
#endif

/* prototypes of internal functions */

double sc_mu (pft *iphase, int newrand, int scaled, double F_max, double F_min);

void cp_direction (direction *dest, direction *source);

double scattering_probability_tot (double mu,
				   atmosphere_struct *atmos, photon_struct *p,
				   int nonHGaer, pft *phase_aer,
				   phase_function_table *phase_cld,
				   phase_function_table *phase_ice,
				   int rismode, int vismode,
				   double *P_isoene, double *P_sct, double refffac, 
                                   int spectral_is);

double get_ksca (atmosphere_struct *atmos, photon_struct *p, int ist);
double get_kscaIS (atmosphere_struct *atmos, photon_struct *p, int ist);

int err_out (char *output, int status);
int mem_err_out (char *output, int line, char *fff, char *file);

double sc_Isotropic_phi();

void cp_photon_struct (photon_struct *dest, photon_struct *source,
		       sample_struct *sample);

double get_phase (pft *iphase, double mu, int scaled);

void mat_mult(double **a, double *b, double *c);
void v_mult  (double *a, double *b, double *c);
void v_sub   (double *a, double *b, double *c);
void v_add   (double *a, double *b, double *c);
void vn_mult (double *a, double *b, double *c);
void v_cross_product (double *a, double *b, double *c);
void v_mult_mu (double *a, double *b, double *c);

double cosd ( double angle );
double sind ( double angle );
double tand ( double angle );
double acosd ( double rad );
double asind ( double rad );
double atand ( double rad );

void destroy_photon (photon_struct *p);

int photon_journey (photon_struct *p, atmosphere_struct *atmos,
		    sample_struct *sample, result_struct *result,
		    elevation_struct *elev, albedo_struct *albedo,
		    surftemp_struct *surftemp,
		    int *absorption, int *source, 
		    int plotpath, int visualize,
		    int nonHGaer,
		    phase_function_table *phase_cld,
		    phase_function_table *phase_ice,
		    pft *phase_aer,
		    float *wvnmlo, float *wvnmhi,
		    float bplkavg, float *refind,
		    float r_earth,
		    int quiet);

void random_reflection (elevation_struct *elev, photon_struct *p);
void random_Lambertian_upward (direction *dir);

int mu_scatter (atmosphere_struct *atmos, photon_struct *p,
		phase_function_table *phase_cld,
		int nonHGaer, pft *phase_aer,
		phase_function_table *phase_ice,
		int scatter_type, double *mu);

int travel_tau (photon_struct *p, atmosphere_struct *atmos,
		sample_struct *sample, result_struct *result,
		elevation_struct *elev, double tau, 
		int absorption, int source, int plotpath, 
		int visualize, float *refind, float r_earth);/*TZ bt*/

struct tnode *addtree_stddev (struct tnode *p, double *loc, double weight);
void treeprint_stddev (struct tnode *p);

double random_cone_mu (double *cotheta);
double random_cone_mu_gauss (double *cotheta);

void atmos_coord  (photon_struct *p, atmosphere_struct *atmos, int *ic, int *jc);

void hitflag (direction *dir);

int set_photon_z (float z, atmosphere_struct *atmos, photon_struct *p);

double get_F (pft *iphase, double mu, int scaled);

double calc_phi_horz (double *dx);

#endif
