/*--------------------------------------------------------------------
 * $Id: uvspec.c,v 1.123 2011/02/28 13:46:59 robert Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include "uvspec.h"
#include "numeric.h"


/* prototypes of internal functions */
void pmesg (char *message, int verbose);



/**************************************************************/
/* The main uvspec function.                                  */
/**************************************************************/

int uvspec (input_struct input, output_struct *output)
{
  int status = 0;
  char function_name[]="uvspec";
  char file_name[]="uvspec.c";


  /**** Check the model input data ****/
  pmesg (" ... calling uvspec_check(), checking model input data\n", input.verbose);
  status = uvspec_check(input); /* in this file: uvspec.c */
  if (status!=0) {
    fprintf (stderr, "%d error(s) in uvspec input-file, aborting\n", abs(status));
    return status;
  }


  /**** Set wavelength grid for the transmittance calculation ****/
  pmesg (" ... calling setup_wlgrid(), generating model wavelength grid\n", input.verbose);
  status = setup_wlgrid (input, output);  /* in ancillary.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up wavelength grid in %s (%s)\n", status, function_name, file_name);
    return status;
  } 


  /**** Setup sza and phi0 ****/
  pmesg (" ... calling setup_sza(), generating model solar zenith and azimuth\n", input.verbose);
  status = setup_sza (input, output); /* in sza.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up sza and phi0 in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Independent pixel stuff ****/
  pmesg (" ... calling setup_ipa(), generating independent pixels\n", input.verbose);
  status = setup_ipa (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up IPA properties in %s (%s)\n", status, function_name, file_name);
    return status;
  }

  
  /**** get surface altitude ****/
  pmesg (" ... calling setup_altitude(), determine surface elevation \n", input.verbose);
  status = setup_altitude (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d determining surface altitude in %s (%s) \n", status, function_name, file_name);
    return status;
  }


  /**** Water cloud stuff (1D) ****/
  pmesg (" ... calling setup_wcloud(), generating water clouds\n", input.verbose);
  status = setup_wcloud (input, output); /* in cloud.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up water cloud properties in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Ice cloud stuff (1D) ****/
  pmesg (" ... calling setup_icloud(), generating ice clouds\n", input.verbose);
  status = setup_icloud (input, output); /* in cloud.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up ice cloud properties in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Cloudless sky atmosphere ****/
  pmesg (" ... calling setup_atmosphere(), generating model atmosphere\n", input.verbose);
  status = setup_atmosphere (input, output); /* in atmosphere.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up atmosphere in %s (%s)\n", status, function_name, file_name);
    return status;
  }  


  /**** Setup temperature at computing levels ****/
  pmesg (" ... calling setup_temperature(), generating model temperature profile\n", input.verbose);
  status = setup_temperature (input, output); /* in atmosphere.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up temperature in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Read cross section files, like O3, BrO, OclO, NO2, etc. ****/
  pmesg (" ... calling setup_crs(), reading cross section files\n", input.verbose);
  status = setup_crs (input, output); /* in molecular.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up absorption cross sections in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Calculate Rayleigh scattering cross section ****/
  pmesg (" ... calling setup_rayleigh(), calculating Rayleigh scattering\n", input.verbose);
  status = setup_rayleigh (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up Rayleigh cross sections in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Calculate Raman scattering cross section ****/
  pmesg (" ... calling setup_raman(), calculating Raman scattering\n", input.verbose);
  status = setup_raman (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up Raman cross sections in %s (%s)\n", status, function_name, file_name);
    return status;
  }

  /****  Optical properties of trace gases ****/
  pmesg (" ... calling setup_gases(), generating optical properties of trace gases\n", input.verbose);
  status = setup_gases (input, output);  /* in atmosphere.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up optical properties of trace gases in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /****   Aerosol stuff   ****/
  pmesg (" ... calling setup_aerosols(), generating aerosols\n", input.verbose);
  status = setup_aerosol (input, output); /* in aerosol.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up aerosol properties in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** read 3D water and ice clouds for MYSTIC ****/
  pmesg (" ... calling setup_atmosphere3D(), generating 3D atmosphere\n", input.verbose);
  status = setup_cloud3D (input, output); /* in cloud3d.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up 3D atmosphere in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Redistribute profiles to required vertical resolution ****/
  pmesg (" ... calling setup_redistribute(), redistributing vertical profiles\n", input.verbose);
  status = setup_redistribute (input, output); /* in redistribute.c */
  if (status!=0) {
    fprintf (stderr, "Error %d doing vertical redistribution of optical properties in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** 3D atmosphere for MYSTIC - to be called after redistribution! ****/
  pmesg (" ... calling setup_atmosphere3D(), generating 3D atmosphere\n", input.verbose);
  status = setup_atmosphere3D (input, output); /* in cloud3d.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up 3D atmosphere in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /****  Read extraterrestrial spectrum and correct for Earth-Sun distance ****/
  pmesg (" ... calling setup_extraterrestrial(), reading extraterrestrial spectrum\n", input.verbose);
  status = setup_extraterrestrial (input, output); /* in extraterrestrial.c */
  if (status!=0) {
    fprintf (stderr, "Error %d setting up extraterrestrial irradiance in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Setup wavelength dependent albedo ****/
  pmesg (" ... calling setup_albedo(), generating surface albedo\n", input.verbose);
  status = setup_albedo (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d setting up wavelength dependent albedo in %s (%s)\n", status, function_name, file_name);
    return status;
  }
  

  /****     Solve RTE     ****/
  pmesg (" ... calling solve_rte(), solving RTE\n", input.verbose);
  status = solve_rte (input, output);
  if (status!=0) {
    fprintf (stderr, "Error %d solving RTE in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Interpolate transmittance to high resolution wavelength grid ****/
  pmesg (" ... calling interpolate_transmittance(), interpolating transmittance\n", input.verbose);
  status = interpolate_transmittance (input, output); /* in ancillary.c */
  if (status!=0) {
    fprintf (stderr, "Error %d interpolating model output to hires wavelength grid in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Multiply with extraterrestrial irradiance ****/
  pmesg (" ... calling multiply_extraterrestrial()\n", input.verbose);
  status = multiply_extraterrestrial (input, output); /* in ancillary.c */
  if (status!=0) {
    fprintf (stderr, "Error %d multiplying with extraterrestrial irradiance in %s (%s)\n", status, function_name, file_name);
    return status;
  }


  /**** Convolve with slit function ****/
  if (input.convolve) {
    pmesg (" ... calling convolve(), convoluting with slit function\n", input.verbose);
    status = convolve (input, output); /* in ancillary.c */
    if (status!=0) {
      fprintf (stderr, "Error %d convolving hires spectrum with slit function in %s (%s)\n", status, function_name, file_name);
      return status;
    }
  }

  /**** Finally interpolate to output grid ****/
  if (input.spline) {
    pmesg (" ... calling spline_interpolate(), interpolating to output grid\n", input.verbose);
    status = spline_interpolate (input, output); /* in ancillary.c */
    if (status!=0) {
      fprintf (stderr, "Error %d interpolating spectrum to output grid in %s (%s)\n", status, function_name, file_name);
      return status;
    }
  }

  
  /**** Process 3D output (wavelength integration, ...) ****/
  if (output->mc.sample.passback3D) {
    pmesg (" ... calling processing3D(), post processing of 3D fields\n", input.verbose);
    status = processing3D (input, output);
    if (status!=0) {
      fprintf (stderr, "Error %d integrating 3D output over wavelength in %s (%s)\n", status, function_name, file_name);
      return status;
    }
  }
 

  /**** Process 1D output (wavelength integration, ...) ****/
  pmesg (" ... calling processing1D(), post processing\n", input.verbose);
  status = processing1D (input, output);   /* in ancillary.c */
  if (status!=0) {
    fprintf (stderr, "Error %d processing 1D output in %s (%s)\n", status, function_name, file_name);
    return status;
  }

  return 0;
}



/**************************************************************/
/* Check the model input for consistency.                     */
/**************************************************************/

int uvspec_check (input_struct Input)
{

#define EPSILON 1.E-6        /*relative rounding error for floats */

  int i=0, status=0;
  int s1=0, s2=0;
  FILE *f=NULL;
  int lev=0;
  float dz=0;

  char function_name[]="uvspec_check";
  char file_name[]="uvspec.c";

  /* ***FIXME*** Still need to check (also in uvspec function):
     aerosol_haze
     aerosol_scale_ssa
     aerosol_set_gg
     aerosol_set_ssa
     aerosol_season
     aerosol_visibility
     aerosol_vulcan
     angstrom
     atmosphere_file
     data_files_path
     nstr
     rte_solver
     solar_file
     sza
     sza_file
     umu
     wc_file
     wc_set_gg
     wc_set_ssa
     wc_layer
     wvn
   */


  /* check latitude and longitude */
  /* position inside normal longitude latitude boundaries ? */
  if ( fabs (Input.latitude - NOT_DEFINED_FLOAT)  > EPSILON && Input.latitude  < - 90.0) {
    fprintf (stderr, "Error in %s (%s), latitude is %7.2f < -90 S\n", function_name, file_name, Input.latitude);
    return -1;
  }
  if ( fabs (Input.latitude - NOT_DEFINED_FLOAT)  > EPSILON && Input.latitude  >   90.0) {
    fprintf (stderr, "Error in %s (%s), latitude is %7.2f > +90 N\n", function_name, file_name, Input.latitude);
    return -1;
  }
  if ( fabs (Input.longitude - NOT_DEFINED_FLOAT) > EPSILON && Input.longitude < -360.0) {
    fprintf (stderr, "Error in %s (%s), longitude is %7.2f < -360 W\n", function_name, file_name, Input.longitude);
    return -1;
  }
  if ( fabs (Input.longitude - NOT_DEFINED_FLOAT) > EPSILON && Input.longitude >  360.0) {
    fprintf (stderr, "Error in %s (%s), longitude is %7.2f > +360 E\n", function_name, file_name, Input.longitude);
    return -1;
  }

  /* time check */
  if ( Input.UTC.tm_year != NOT_DEFINED_INTEGER && (Input.UTC.tm_year < 1902.-1900. || 2038.-1900. <= Input.UTC.tm_year) ) {
    fprintf (stderr, "Error, sorry the mktime-function of C does only work for the time range\n");
    fprintf (stderr, "       1902 ... 2038. If anybody has a function with broader range, \n");
    fprintf (stderr, "       please email it to the libRadtran programmers. Thanks. \n");
    return -1;
  } 

  /* Albedo checks */
  /* Albedo value range for direct input */
  if (Input.alb.albedo > 1.0) {
    fprintf (stderr, "Error in %s (%s), surface albedo %f > 1 \n", function_name, file_name, Input.alb.albedo);
    status--;
  }
  else if (Input.alb.albedo < 0.0 && Input.alb.source == ALBEDO_CONSTANT) {
    fprintf (stderr, "Error in %s (%s), surface albedo %f < 0\n", function_name, file_name, Input.alb.albedo);
    status--;
  }

  if ( Input.alb.surf_type_map == TRUE )
    if ( Input.alb.source != ALBEDO_USER_LIBRARY && Input.alb.source != ALBEDO_IGBP_LIBRARY) {
      fprintf (stderr, "Error, surface_type_map was specified, but also another albedo information in the input file.\n");
      switch(Input.alb.source) {
      case ALBEDO_CONSTANT:
        fprintf (stderr, "       constant albedo\n");
        break;
      case ALBEDO_FROM_ALBEDO_MAP:
        fprintf (stderr, "       albedo_map \n");
        break;
      case ALBEDO_FROM_EMISSIVITY_MAP:
        fprintf (stderr, "       emissivity_map \n");
        break;
      case ALBEDO_FROM_ALBEDO_FILE:
        fprintf (stderr, "       albedo_file \n");
        break;
      case ALBEDO_USER_LIBRARY:
      case ALBEDO_IGBP_LIBRARY:
        /* does never happen */
        break;
      default:
        fprintf (stderr, "Error: unsupported albedo source in %s (%s)\n", function_name, file_name);
        return -1;
      }
      fprintf (stderr, "       please specify only one albedo information in the input file! \n");
      status--;
    }

  if (Input.alb.source == ALBEDO_FROM_EMISSIVITY_MAP && Input.source != SRC_THERMAL) {
    fprintf (stderr, "Error, emissivity_map only possible for thermal simulations in %s (%s)\n", function_name, file_name);
    fprintf (stderr, "       please use albedo_map instead! \n");
    status--;
  }

  /* Day of year checks */

  if (Input.UTC.tm_yday > 366) {
    fprintf (stderr, "Error in %s (%s), day of year %d > 366\n", function_name, file_name, Input.UTC.tm_yday);
    status--;
  }
  else if (Input.UTC.tm_yday < 0 && Input.UTC.tm_yday != -999) {
    fprintf (stderr, "Error in %s (%s), day of year %d < 0 \n", function_name, file_name, Input.UTC.tm_yday);
    status--;
  }

  /* ECMWF data + time checks */
  if ((Input.UTC.tm_mday <= 0) && 
      (Input.atm.rs_source == RS_FROM_ECMWF     || 
       Input.wc.source     == CLOUDS_FROM_ECMWF ||
       Input.ic.source     == CLOUDS_FROM_ECMWF   )) {
    fprintf (stderr, "Error, if you use ECMWF data, it is nessesary to specify also the time;\n");
    fprintf (stderr, "       please use the 'time' option in the input file\n");
    status--;
  }


  /* Ozone column checks */

  if (Input.atm.column[MOL_O3] < 0.0 && Input.atm.column[MOL_O3] != -999) {
    fprintf (stderr, "Error in %s (%s), ozone column %f < 0\n", function_name, file_name, Input.atm.column[MOL_O3]);
    status--;
  }


  /* Precipitable water checks */

  if (Input.precip_water < 0.0 && Input.precip_water != -999) {
    fprintf (stderr, "Error in %s (%s), precipitable water %f < 0\n", function_name, file_name, Input.precip_water);
    status--;
  }

  if (Input.precip_water > 0.0 && Input.mixing_ratio[MX_H2O]>0) {
    fprintf (stderr, "Error in %s (%s), it does not make sense to define both h2o_mixing_ratio and h2o_precip\n", function_name, file_name);
    status--;
  }


  /* Aerosol checks */

  /* if any of the following four variables is set, all four must be set; */
  /* otherwise the information is incomplete!                             */
  if (Input.aer.seasn>0 || Input.aer.vulcan>0 || Input.aer.haze>0 || Input.aer.visibility>0) {

    if (Input.aer.seasn<=0) {
      fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_season or aerosol_default!\n", function_name, file_name);
      status--;
    }

    if (Input.aer.vulcan<=0) {
      fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_vulcan or aerosol_default!\n", function_name, file_name);
      status--;
    }

    if (Input.aer.haze<=0) {
      fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_haze or aerosol_default!\n", function_name, file_name);
      status--;
    }

    if (Input.aer.visibility<=0) {
      fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_visibility or aerosol_default!\n", function_name, file_name);
      status--;
    }
  }
  
  /* aerosol_default does not work with polarization */
  if (Input.aer.seasn>0 && Input.aer.n_species == NOT_DEFINED_INTEGER  && 
      ((Input.rte.solver == POLRADTRAN &&
        Input.rte.pol_nstokes>1) || 
       Input.rte.mc.polarisation == 1 )) {
    fprintf (stderr, "Error in %s (%s), default aerosol is not implemented for polarized radiance calculations!\nPlease use option aerosol_species_file (e.g., OPAC aerosol).\n", function_name, file_name);
    status--;
    }


  /* if no standard aerosol is defined, we don't want any of the aerosol varibles set */
  if (!Input.aer.standard && Input.aer.spec) {  
    fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_default or each\n", function_name, file_name);
    fprintf (stderr, "of aerosol_vulcan, aerosol_vulcan, aerosol_haze, and aerosol_visibility!\n"); 
    status--;
  }


  if ((Input.aer.seasn < 0 && Input.aer.seasn!=-999) || Input.aer.seasn >= AEROSOL_SEASON_NN) {
    fprintf (stderr, "Error in %s (%s), aerosol season out of range.\n", function_name, file_name);
    status--;
  }

  if ((Input.aer.haze < 0 && Input.aer.haze!=-999) || Input.aer.haze >= AEROSOL_HAZE_NN) {
    fprintf (stderr, "Error in %s (%s), aerosol haze out of range.\n", function_name, file_name);
    status--;
  }

  if ((Input.aer.vulcan < 0 && Input.aer.vulcan!=-999) || Input.aer.vulcan >= AEROSOL_VULCAN_NN) {
    fprintf (stderr, "Error in %s (%s), aerosol vulcan out of range.\n", function_name, file_name);
    status--;
  }
    

  s1=strlen(Input.aer.filename[FN_AER_SIZ]);
  s2=strlen(Input.aer.filename[FN_AER_REF]);

  if (s1>0 && (s2==0 && Input.aer.re==0)) {
    fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_refrac_index \n", function_name, file_name);
    fprintf (stderr, "or aerosol_refrac_file or omit the size distribution\n");
    status--;
  }

  if (s1==0 && (s2>0 || Input.aer.re>0)) {
    fprintf (stderr, "Error in %s (%s), aerosol information incomplete; specify aerosol_sizedist_file\n", function_name, file_name);
    fprintf (stderr, "or omit the refractive index!\n");
    status--;
  }

  /* Cloud checks */
  
  if (Input.wc.tau>=0.0 && Input.wc.tau550>=0.0) {
    fprintf (stderr, "Error in %s (%s), it doesn't make sense to specify both wc_set_tau and wc_set_tau550;\n", function_name, file_name);
    fprintf (stderr, "please remove either one and retry!\n");
    status--;
  }

  if (Input.ic.tau>=0.0 && Input.ic.tau550>=0.0) {
    fprintf (stderr, "Error in %s (%s), it does not make sense to specify both ic_set_tau and ic_set_tau550;\n", function_name, file_name);
    fprintf (stderr, "please remove either one and retry!\n");
    status--;
  }

  if (strlen(Input.filename[FN_WCLOUD])>0 && strlen(Input.filename[FN_WCLOUD_FILES])>0) {
    fprintf (stderr, "Error in %s (%s), it does not make sense to specify both wc_file and wc_files;\n", function_name, file_name);
    fprintf (stderr, "please remove either one and retry!\n");
    status--;
  }
  
  /* water cloud and polarization works only for mie files */
  if (Input.wc.source != NO_CLOUDS && 
      (Input.wc.properties != WCPROP_MIE && Input.wc.properties != WCPROP_FILE) &&
      ((Input.rte.solver == POLRADTRAN &&
        Input.rte.pol_nstokes>1) || 
       Input.rte.mc.polarisation == 1 )) {
    fprintf (stderr, "Error in %s (%s), for polarized radiance calculations in water clouds \n you have to use wc_properties mie or \n  specify your own optical properties data file!\n", function_name, file_name);
    status--;
  }

  /* need to specify a cloud for the given cloud properties */
  if ((Input.ic.properties != ICPROP_NONE && Input.ic.properties != ICPROP_EXPLICIT) &&
      (strlen(Input.filename[FN_ICLOUD])==0 && strlen(Input.filename[FN_ICLOUD_IPA])==0 && strlen(Input.rte.mc.filename[FN_MC_ICLOUD])==0)) {/*TZ*/
    fprintf (stderr, "Error, ic_properties only makes sense together with 'ic_file'\n");
    fprintf (stderr, "or 'ic_ipa_files' or 'mc_icloud_file'!\n");
    status--;
  }
  
  if (strlen(Input.filename[FN_ICLOUD_PROP])>0 && !Input.ic.layer && strlen(Input.rte.mc.filename[FN_MC_ICLOUD])==0) {  /*TZ*/
    fprintf (stderr, "Error, ic_properties requires optical properties defined per layer for 1D clouds.\n");
    fprintf (stderr, "Please specify ic_layer and retry if that is what you mean!\n");
    status--;
  }
  
  /* ice cloud and polarization not yet included in distribution, only option is to specify an*/ 
  /* optprop file */
  if (Input.ic.source != NO_CLOUDS && Input.ic.properties != ICPROP_FILE && Input.ic.properties != ICPROP_HEY && 
      ((Input.rte.solver == POLRADTRAN &&
        Input.rte.pol_nstokes>1) || 
       Input.rte.mc.polarisation == 1 )) {
    fprintf (stderr, "Error in %s (%s), for polarized radiance calculations with ice clouds \n",
             function_name, file_name);
    fprintf (stderr, "you have to use ic_properties HEY or specify an optical properties file \n");
    fprintf (stderr, "using ic_properties filename!\n");
    status--;
  }
  
  

  /* RTE checks */

  switch (Input.rte.solver) {
  case SPSDISORT:
    if (Input.source==SRC_THERMAL) {
      fprintf (stderr, "Error, spsdisort does not include thermal emission\n");
      status--;
    }

    break;

  case SDISORT:

    if (Input.source==SRC_THERMAL) {
      fprintf (stderr, "Error, sdisort does not include thermal emission\n");
      status--;
    }

    if (Input.rte.nscat < 0 || Input.rte.nscat > 2) {
      fprintf (stderr, "Error, nscat %d is out of bounds.\n", Input.rte.nscat);
      fprintf (stderr, "Only 1 or 2 is allowed in combination with the sdisort solver.\n");
      status--;
    }

    if (Input.rte.nrefrac < 0 || Input.rte.nrefrac > 2) {
      fprintf (stderr, "Error, nrefrac = %d is out of bounds\n", Input.rte.nrefrac);
      status--;
    }
    
    break;

  case TWOSTR:
  case TWOSTRPP:
  case RODENTS:
  case TWOSTREBE:

    if (Input.rte.nprndis>0)
      for (i=0;i<Input.rte.nprndis;i++)
	if (!(Input.rte.prndis[i] == 1 || Input.rte.prndis[i] == 2)) { 
	  fprintf (stderr, "Error, twostr can only handle prndis 1 and/or 2.\n");
	  status--;
	}

    break;

  case SSLIDAR:
  case DISORT:
  case SUCORD:
  case MONTECARLO:
  case POLRADTRAN:
  case DISORT2:
  case QDISORT:
  case TZS:
  case SSS:
  case SSSI:
  case NULLSOLVER:
  case CTWOSTR:
  case CDISORT:
    break;

  default:
    fprintf (stderr, "Error, unknown rte_solver %d in uvspec_check\n", Input.rte.solver);
    return -1;
  }


  if (Input.rte.mc.absorption==MCFORWARD_ABS_EMISSION && Input.source != SRC_THERMAL) {
    fprintf (stderr, "Error, mc_emission does only make sense with \"source thermal\"\n");
    status--;
  }

  /* either specify a sensor direction or calculate surface-parallel */
  if (Input.rte.mc.surfaceparallel && Input.rte.mc.sensordirection) {
    fprintf (stderr, "Error, mc_surfaceparallel together with  mc_sensordirection\n");
    fprintf (stderr, "does not make sense. Please remove one of them!\n");
    status--;
  }

  /* sensor direction does only work in backward mode */
  if (Input.rte.mc.sensordirection && !Input.rte.mc.backward.yes) {
    fprintf (stderr, "Error, mc_sensordirection only possible in backward mode!\n");
    status--;
  }

  /* sensor position does only work in backward mode */
  if (Input.rte.mc.sensorposition && !Input.rte.mc.backward.yes) {
    fprintf (stderr, "Error, mc_sensorposition only possible in backward mode!\n");
    status--;
  }

  /* check if absorption/emission was consistently defined in forward/backward mode; */
  /* if both are defined, unneccesarily large amounts of memory might be allocated.  */
  if (Input.rte.mc.absorption && Input.rte.mc.backward.absorption) {
    fprintf (stderr, "Error, specified both forward and backward calculation\n");
    fprintf (stderr, "of absorption/emission/actinic flux. Please choose either\n");
    fprintf (stderr, "forward (mc_absorption, ...) or backward (mc_backward_output abs).\n");
    fprintf (stderr, "The programmers are sorry for any inconvenience this might have caused ...\n");
    status--;
  }
    

  /* check if z_atm_forced are sorted in ascending (descending) order */
  /* the order of the levels is reversed just direct after the input  */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER) {
    for (i=1; i<Input.atm.nz_atm_forced_sea; i++) {
      if (Input.atm.z_atm_forced_sea[i] - Input.atm.z_atm_forced_sea[i-1] >= 0 ) {
        fprintf (stderr, "Error, 'atmosphere_zgrid' levels not sorted in ascending order\n");
        fprintf (stderr, "       checking atmosphere_zgrid-levels in %s (%s)\n", function_name, file_name);
        for (lev=0; lev<Input.atm.nz_atm_forced_sea; lev++)
          fprintf (stderr, "    zgrid[%3d] = %8.3f\n",lev,Input.atm.z_atm_forced_sea[Input.atm.nz_atm_forced_sea-1-lev]);
        status--;
      }
    }
  }

  /* forbidden combination z_atm_forced and altitude with 2 arguments */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER && Input.alt.altitude_dz > 0.0 ) {
    fprintf (stderr, "Error, combination 'atmosphere_zgrid' and 'altitude with second argument' is not allowed\n");
    status--;
  }
  /* forbidden combination z_atm_forced and molecular_tau_file */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER && strlen(Input.filename[FN_MOLABS]) > 0) {
    fprintf (stderr, "Error, combination 'atmosphere_zgrid' and 'molecular_tau_file' is not allowed\n");
    status--;
  }
  /* forbidden combination z_atm_forced and rayleigh_tau_file */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER && strlen(Input.filename[FN_RAYLEIGH]) > 0) {
    fprintf (stderr, "Error, combination 'atmosphere_zgrid' and 'rayleigh_tau_file' is not allowed\n");
    status--;
  }

  /* check if z_atm_forced are sorted in ascending (descending) order */
  /* the order of the levels is reversed just direct after the input  */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER) {
    for (i=1; i<Input.atm.nz_atm_forced_sea; i++) {
      if (Input.atm.z_atm_forced_sea[i] - Input.atm.z_atm_forced_sea[i-1] >= 0 ) {
        fprintf (stderr, "Error, 'atmosphere_zgrid' levels not sorted in ascending order\n");
        fprintf (stderr, "       checking atmosphere_zgrid-levels in %s (%s)\n", function_name, file_name);
        for (lev=0; lev<Input.atm.nz_atm_forced_sea; lev++)
          fprintf (stderr, "    zgrid[%3d] = %8.3f\n",lev,Input.atm.z_atm_forced_sea[Input.atm.nz_atm_forced_sea-1-lev]);
        status--;
      }
    }
  }

  /* forbidden combination z_atm_forced and altitude with 2 arguments */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER && Input.alt.altitude_dz > 0.0 ) {
    fprintf (stderr, "Error, combination 'atmosphere_zgrid' and 'altitude with second argument' is not allowed\n");
    status--;
  }
  /* forbidden combination z_atm_forced and molecular_tau_file */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER && strlen(Input.filename[FN_MOLABS]) > 0) {
    fprintf (stderr, "Error, combination 'atmosphere_zgrid' and 'molecular_tau_file' is not allowed\n");
    status--;
  }
  /* forbidden combination z_atm_forced and rayleigh_tau_file */
  if (Input.atm.nz_atm_forced_sea != NOT_DEFINED_INTEGER && strlen(Input.filename[FN_RAYLEIGH]) > 0) {
    fprintf (stderr, "Error, combination 'atmosphere_zgrid' and 'rayleigh_tau_file' is not allowed\n");
    status--;
  }

  /* check if zout is sorted in ascending order */
  switch (Input.atm.zout_source) {
  case OUTLEVEL_ZOUT_ABOVE_SUR:
    for (i=1; i<Input.atm.nzout; i++) {
      if (Input.atm.zout_sur[i] - Input.atm.zout_sur[i-1] <= 0 ) {
        fprintf (stderr, "Error, zout (above surface) not sorted in ascending order\n");
        fprintf (stderr, "       checking zout_sur-level in %s (%s)\n", function_name, file_name);
        for (lev=0; lev<Input.atm.nzout; lev++)
          fprintf (stderr, "    zout[%3d] = %8.3f\n",lev,Input.atm.zout_sur[lev]);
        status--;
      }
    }
    break;
  case OUTLEVEL_ZOUT_ABOVE_SEA:
    for (i=1; i<Input.atm.nzout; i++) {
      if (Input.atm.zout_sea[i] - Input.atm.zout_sea[i-1] <= 0 ) {
        fprintf (stderr, "Error, zout (above sea level) not sorted in ascending order\n");
        fprintf (stderr, "       checking zout_sea-level in %s (%s)\n", function_name, file_name);
        for (lev=0; lev<Input.atm.nzout; lev++)
          fprintf (stderr, "    zout[%3d] = %8.3f\n",lev,Input.atm.zout_sea[lev]);
        status--;
      }
    }
    break;
  case OUTLEVEL_PRESS:
    for (i=1; i<Input.atm.nzout; i++) {
      if ( Input.atm.press_zout[i] - Input.atm.press_zout[i-1] >= 0
           && Input.atm.press_zout[i-1] != ZOUT_SURFACE
           && Input.atm.press_zout[i]   != ZOUT_TOA
           && !( Input.atm.press_zout[i] == ZOUT_CPT && i==Input.atm.nzout-1 ) ) {
        if ( (Input.atm.press_zout[i] == ZOUT_CPT || Input.atm.press_zout[i-1] == ZOUT_CPT) && i!=Input.atm.nzout-1) {
          fprintf (stderr, "Error, outputlevel 'CPT' must be the last level.\n");
        }
        else {
          fprintf (stderr, "Error, outputlevel 'pressure_out' not sorted in descending order %f > %f \n",
                                                    Input.atm.press_zout[i],Input.atm.press_zout[i-1] );
        }
        fprintf (stderr, "       checking pressure_out-level in %s (%s)\n", function_name, file_name);
        for (lev=0; lev<Input.atm.nzout; lev++)
          fprintf (stderr, "    pressure_out[%3d] = %8.3f\n",lev,Input.atm.press_zout[lev]);
        status--;
      }
    }
    break;
  case OUTLEVEL_ATM_LEVELS:
  case OUTLEVEL_ALL_LEVELS:
  case OUTLEVEL_MODEL_LEVELS:
  case OUTLEVEL_MODEL_LAYERS:
  case OUTLEVEL_MODEL_LEVELS_AND_LAYERS:
    /* no checks here */
    break;
  default:
    fprintf (stderr, "Error, unknown output level type %d (function %s in %s) \n", Input.atm.zout_source, __func__, __FILE__ );
    fprintf (stderr, "This is a program bug, please report this bug to the programmers and include the input file. Thanx.\n");
    return -1;
  }

  /* transmittance_wl_file and thermal_bands_file are mutually exclusive */
  if ((strlen(Input.filename[FN_WLTRANS])>0) &&
      (strlen(Input.filename[FN_WLBANDS])>0)) {
    fprintf (stderr, "Error, it does not make sense to specify both transmittance_wl_file and thermal_bands_file!\n");
    status--;
  }

  /* solar_file and thermal_bands_file are mutually exclusive */
  if ((strlen(Input.filename[FN_EXTRATERRESTRIAL])>0) &&
      (strlen(Input.filename[FN_WLBANDS])>0)) {
    fprintf (stderr, "Error, it does not make sense to specify both\n");
    fprintf (stderr, "       solar_file and thermal_bands_file!\n");
    status--;
  }

  /* Need to define a wavelength range if no correlated_k and no transmittance_wl_file */
  if ((Input.wl.start<0 || Input.wl.end<0) && 
      (Input.ck_scheme == CK_NONE || Input.ck_scheme == CK_RAMAN || Input.ck_scheme == CK_LOWTRAN) &&
      (strlen(Input.filename[FN_WLTRANS])==0) &&
      (strlen(Input.filename[FN_WLBANDS])==0) &&
      (strlen(Input.filename[FN_MOLABS])==0)) {
    fprintf (stderr, "Error, need to define a wavelength range with 'wvn lower upper'!\n");
    status--;
  }

  /* wvn_index only makes sense with explicitely defined wavelength indices */
  if ((Input.wl.start_index>0 || Input.wl.end_index>0) && 
      ((Input.ck_scheme == CK_NONE || Input.ck_scheme == CK_RAMAN || Input.ck_scheme == CK_LOWTRAN) &&
       (strlen(Input.filename[FN_WLTRANS])==0) &&
       (strlen(Input.filename[FN_WLBANDS])==0))) {
    fprintf (stderr, "Error, wvn_index can only be used together with\n");
    fprintf (stderr, "explicitely defined wavelength grids!\n");
    status--;
  }

  /* wvn and wvn_index are mutually exclusive */
  if ((Input.wl.start_index>0 || Input.wl.end_index>0) && 
      (Input.wl.start>0 || Input.wl.end>0)) {
    fprintf (stderr, "Error, it does not make sense to define both 'wvn' and 'wvn_index'!\n");
    status--;
  }

  if (Input.spline && !(strlen(Input.filename[FN_SPLINE])>0)) {
    if (Input.wl.start>0 && Input.wl.start > Input.spline_lambda_0)  {
      fprintf (stderr, "Error, spline wavelength %f is smaller than wvn %f\n",
	       Input.spline_lambda_0, Input.wl.start);
      status--;
    }
    if (Input.wl.end>0 && Input.wl.end < Input.spline_lambda_1)  {
      fprintf (stderr, "Error, spline wavelength %f is greater than wvn %f\n",
	       Input.spline_lambda_1, Input.wl.end);
      status--;
    }
  }

  /* Check if slit_function file exists */
  if (Input.convolve) {
    if ((f = fopen(Input.filename[FN_SLITFUNCTION], "r")) == NULL) {
      fprintf (stderr, "Error, cannot read slit function file %s\n", Input.filename[FN_SLITFUNCTION]);
      status--;
    }
    else 
      fclose(f);
  }

  
  /* Cloud checks */
  if (Input.wc_cloudcover != NOT_DEFINED_FLOAT) {
    if (Input.wc_cloudcover < 0.0 || Input.wc_cloudcover > 1.0) {
      fprintf (stderr, "Error, wc_cloudcover %f not between 0 and 1\n", Input.wc_cloudcover);
      status--;
    }
    if (strlen(Input.filename[FN_WCLOUD])==0) {
      fprintf (stderr, "Error, wc_cloudcover must be accompanied by a cloud file wc_file\n");
      status--;
    }
  }

  if (Input.ic_cloudcover != NOT_DEFINED_FLOAT) {
    if (Input.ic_cloudcover < 0.0 || Input.ic_cloudcover > 1.0) {
      fprintf (stderr, "Error, ic_cloudcover %f not between 0 and 1\n", Input.ic_cloudcover);
      status--;
    }
    if (strlen(Input.filename[FN_ICLOUD])==0) {
      fprintf (stderr, "Error, ic_cloudcover must be accompanied by a cloud file ic_file\n");
      status--;
    }
  }

  if (Input.wc_cloudcover != NOT_DEFINED_FLOAT && Input.ic_cloudcover != NOT_DEFINED_FLOAT) {
    if (Input.wc_cloudcover != Input.ic_cloudcover) {
      fprintf (stderr, "Error, wc_cloudcover and ic_cloudcover must be equal!");
      status--;
    }
  }


  
  /* ck and libnetcdf */ 
#if HAVE_LIBNETCDF
#else
  if (Input.ck_scheme == CK_KATO || Input.ck_scheme == CK_KATO2 || Input.ck_scheme == CK_KATO2_96 || 
      Input.ck_scheme == CK_GENERIC) {
    fprintf (stderr, "***********************************************************************\n");
    fprintf (stderr, "* Error, you have built uvspec without libnetcdf and hence may not    *\n");
    fprintf (stderr, "* use the 'correlated_k kato' options. Please get netcdf and rebuild. *\n");
    fprintf (stderr, "***********************************************************************\n");
    status--;
  }
#endif


  /* The MYSTIC solver */ 
#if HAVE_MYSTIC
#else
  if (Input.rte.solver == MONTECARLO) {
    fprintf (stderr, "*********************************************************************\n");
    fprintf (stderr, "* Error, you have built uvspec without the three-dimensional MYSTIC *\n");
    fprintf (stderr, "* module. MYSTIC is currently not released to the public. If you    *\n");
    fprintf (stderr, "* have an interest in three-dimensional calculations and if you     *\n");
    fprintf (stderr, "* have a reasonable project in mind, please contact                 *\n");
    fprintf (stderr, "* bernhard.mayer@dlr.de.                                            *\n");
    fprintf (stderr, "*********************************************************************\n");
    status--;
  }
#endif


  /* Fu and Liou */ 
#if HAVE_FULIOU
#else
  if (Input.ck_scheme == CK_FU) {
    fprintf (stderr, "**********************************************************************\n");
    fprintf (stderr, "* Error, you have built uvspec without support for the Fu and Liou   *\n");
    fprintf (stderr, "* parameterization. If you are interested, please contact the        *\n");
    fprintf (stderr, "* libRadtran authors.                                                *\n");
    fprintf (stderr, "**********************************************************************\n");
    status--;
  }
#endif
  

  if ((Input.ic.properties==ICPROP_YANG || Input.ic.properties==ICPROP_KEY) && Input.ic.habit<0) {
    fprintf (stderr, "Error, unknown ice crystal habit. Check ic_habit!\n");
    status--;
  }


  /* LOWTRAN/SBDART */ 
#if HAVE_LOWTRAN
  if (Input.ck_scheme == CK_LOWTRAN && !Input.quiet) {
    fprintf (stderr, "*************************************************************************\n");
    fprintf (stderr, "* When using the LOWTRAN/SBDART gas absorption for any result that goes *\n");
    fprintf (stderr, "* into a publication, please don't forget to add a reference like:      *\n");
    fprintf (stderr, "* \"Molecular absorption was parameterized with the LOWTRAN band        *\n");
    fprintf (stderr, "*   model (Pierluissi and Peng, 1985), as adopted from the SBDART code  *\n");
    fprintf (stderr, "*   (Ricchiazzi et al., 1998).                                          *\n");
    fprintf (stderr, "*************************************************************************\n");
  }
#else
  if (Input.ck_scheme == CK_LOWTRAN) {
    fprintf (stderr, "***********************************************************************\n");
    fprintf (stderr, "* Error, you have built uvspec without support for the LOWTRAN/SBDART *\n");
    fprintf (stderr, "* gas parameterization. If you are interested, please contact the     *\n");
    fprintf (stderr, "* libRadtran authors.                                                 *\n");
    fprintf (stderr, "***********************************************************************\n");
    status--;
  }
#endif

  /* consistent?  output_unit (W/(m2 nm) or W/(m2 cm-1) or W/(m2 band)) and processing ('output sum' or 'output integrate')*/
  switch(Input.output_unit) {
  case UNIT_PER_NM:
    if (Input.processing == PROCESS_SUM && !Input.quiet) {
      fprintf (stderr, "*** Warning, the result of the radiative transfer calculation is given\n");
      fprintf (stderr, "*** per nano meter: W/(m**2 nm) because 'output per_nm' was chosen.\n");
      fprintf (stderr, "*** In combination with 'output sum', the sum of the spectral results is only equal\n");
      fprintf (stderr, "*** to the integrated result, if the wavelength calculation grid \n");
      fprintf (stderr, "*** (defined by the solar_file) has 1nm steps.\n\n");
    }
    break;
  case UNIT_PER_CMm1:
    if (Input.processing == PROCESS_SUM && !Input.quiet) {
      fprintf (stderr, "*** Warning, the result of the radiative transfer calculation is given\n");
      fprintf (stderr, "*** per wavenumber interval: W/(m**2 cm-1) because 'output per_cm-1' was chosen.\n");
      fprintf (stderr, "*** In combination with 'output sum', the sum of the spectral results is only equal\n");
      fprintf (stderr, "*** to the integrated result, if the wavelength calculation grid (defined by the solar_file)\n");
      fprintf (stderr, "*** has 1cm-1 steps and the bandwidth of each calculation is 1cm-1 (defined by thermal_bandwidth)\n");
    }
    break;
  case UNIT_PER_CK_BAND:
    if (Input.processing == PROCESS_INT) {
      fprintf (stderr, "Error, the result of the radiative transfer calculation is given\n");
      fprintf (stderr, "per wavelength band: W/(m2 band) because 'output per_ck_band' was chosen.\n");
      fprintf (stderr, "This does not make sense in combination with 'output integrate'!\n");
      fprintf (stderr, "Please use 'output sum' instead of 'output integrate' in the input file.\n");
      return -1;
    }
    break;
  case UNIT_NOT_DEFINED:
    /* not defined, that's OK, the user has to know what he is doing */
    break;
  default:
    fprintf (stderr, "Error, unknown output option %d. (function %s in %s) \n", Input.output_unit, __func__, __FILE__ );
    fprintf (stderr, "This is a program bug, please contact the programmers!\n");
    return -1;
  }

  /* Output checks */
  if (Input.heating != NO_HEAT) {
    switch (Input.rte.solver) {
    case NULLSOLVER:
    case TWOSTR:
    case CTWOSTR:
    case TWOSTRPP:
    case RODENTS:
    case TWOSTREBE:
    case DISORT:
    case CDISORT:
    case SDISORT:
    case DISORT2:
    case SPSDISORT:
    case MONTECARLO:
      break;
    case POLRADTRAN:
      if (Input.heating == HEAT_LOCAL) {
        fprintf (stderr, "Error, 'heating_rate local' cannot be used with polradtran.\n");
        fprintf (stderr, "Please choose heating_rate layer_cd, layer_fd, or a different rte_solver.\n");
        status--;
      }
      break;
    default:
      fprintf (stderr, "Error, the chosen solver %d does not provide heating rates.\n", Input.rte.solver);
      status--;
    }

    if ( Input.heating == HEAT_LOCAL &&  Input.rte.solver == MONTECARLO ) {
      fprintf (stderr, "Error, combination 'heating_rate local' and 'rte_solver MONTECARLO' \n");
      fprintf (stderr, "       is not allowed, as local heating rates require additional z-levels.\n");
      fprintf (stderr, "       Please use: 'mc_absorption K_per_day' instead \n");
      status--;
    }

    if (Input.heating == HEAT_LAYER_CD || Input.heating == HEAT_LAYER_FD) {
      if ( Input.atm.nzout < 2 && 
           Input.atm.zout_source != OUTLEVEL_ATM_LEVELS && 
           Input.atm.zout_source != OUTLEVEL_ALL_LEVELS &&
           Input.atm.zout_source != OUTLEVEL_MODEL_LEVELS &&
           Input.atm.zout_source != OUTLEVEL_MODEL_LAYERS &&
           Input.atm.zout_source != OUTLEVEL_MODEL_LEVELS_AND_LAYERS ) {
        fprintf (stderr, "Error, to calculate heating rates, at least two output levels are required.\n");
        fprintf (stderr, "Please use the zout command to specify them!\n");
        status--;
      }
    }
    if (Input.heating == HEAT_LAYER_CD) {

      switch (Input.atm.zout_source) {
      case OUTLEVEL_ZOUT_ABOVE_SUR:
        dz = fabs(Input.atm.zout_sur[1]-Input.atm.zout_sur[0]);
        for (lev=1; lev<Input.atm.nzout-1; lev++)      {
	  if (fabs(dz - (Input.atm.zout_sur[lev+1]-Input.atm.zout_sur[lev]))/dz > 0.1 && !Input.quiet) {
	    fprintf (stderr, "*** Warning, zout layers should be equally thick;\n");
            fprintf (stderr, "*** otherwise output might be inaccurate!\n"); 
            break;
          }
	}
	break;
      case OUTLEVEL_ZOUT_ABOVE_SEA:
        dz = fabs(Input.atm.zout_sea[1]-Input.atm.zout_sea[0]);
        for (lev=1; lev<Input.atm.nzout-1; lev++)      {
	  if (fabs(dz - (Input.atm.zout_sea[lev+1]-Input.atm.zout_sea[lev]))/dz > 0.1 && !Input.quiet) {
	    fprintf (stderr, "*** Warning, zout layers should be equally thick;\n");
            fprintf (stderr, "*** otherwise output might be inaccurate!\n"); 
            break;
          }
	}
	break;
      case OUTLEVEL_PRESS:
      case OUTLEVEL_ATM_LEVELS:
      case OUTLEVEL_ALL_LEVELS:
      case OUTLEVEL_MODEL_LEVELS:
      case OUTLEVEL_MODEL_LAYERS:
      case OUTLEVEL_MODEL_LEVELS_AND_LAYERS:
        /* no warning in this case, as it is not possible at this moment jet */
	break;
      default:
        fprintf (stderr, "Error, unknown output level type %d. (function %s in %s) \n", Input.atm.zout_source, __func__, __FILE__);
        fprintf (stderr, "This is a program bug, please report this bug to the programmers and include the input file. Thanx.\n");
        return -1;
      }
    }
  }

  if (Input.rte.mc.backward.yes && Input.rte.solver == MONTECARLO)
    if (Input.rte.mc.absorption==MCFORWARD_ABS_HEATING && Input.source == SRC_SOLAR) {
      fprintf (stderr, "Error, solar heating rates not implemented in backward Monte Carlo mode.\n");
	fprintf (stderr, "This could be done but probably forward is more efficient than backward\n");
	fprintf (stderr, "because photons usually contribute to many layers (and heating rates are\n");
	fprintf (stderr, "usually calculated for many layers rather than only one).\n");
	status--;
    }
  




  /* can only simulate wind output, if there is a wind map defined */
  for ( i=0; i<Input.n_output_user; i++ ) 
    if (Input.output_user[i]==OUTPUT_USER_WIND_U || Input.output_user[i]==OUTPUT_USER_WIND_V || Input.output_user[i]==OUTPUT_USER_WIND_W || 
        Input.output_user[i]==OUTPUT_USER_HEAT_AD_X || Input.output_user[i]==OUTPUT_USER_HEAT_AD_Y || Input.output_user[i]==OUTPUT_USER_HEAT_AD_Z || 
        Input.output_user[i]==OUTPUT_USER_HEAT_AD ) {
      if (strlen(Input.filename[FN_ECMWF_WIND_MAP])==0) {
        fprintf (stderr, "Error, in order to get wind or advection information \n");
        fprintf (stderr, "       it is crutial to define an ECMWF_wind_file. \n");
        return -1;
      }
    }
  
  return status;
}





void pmesg (char *message, int verbose)
{
  if (verbose) {
    fprintf (stderr, message);
    fflush (stderr);
  }
}
